/*
   File: ProcessFunctions.cc
   Author: Tom Vance
   Date: 10/21/2010

   This is a collection of functions to deal with stuck runtimes.
   On both terminal services and WEB servers the runtime tends to be left stuck
   running with the need to kill it.  You can use the following functions to see
   what is running and what needs to be killed.

   The first three functions are from Processes and Threads in the Kernal32.dll
   For more information please see: 
   http://msdn.microsoft.com/en-us/library/ms684841(v=VS.85).aspx

   The last function uses a VB Script function thru MS Script Control to do a Query
   of all the Running Processes or to gather information about only one running process.
   For more information about this please see:
   http://msdn.microsoft.com/en-us/library/aa393964(v=VS.85).aspx

   The real reason to get information about a running process is that it is the 
   only way I know to fully identify what is running before you kill it.
   The PID is reused in a rapid manner and randomly so you can not count on one 
   being unique for very long.

   How to use this collection:

   Set procedure to ProcesssFunctions.cc
   p = new ProcessFunctions()
   pid = p.ProcessId()  //This will give you the Process Id of the calling program

   The idea is to store the pid somewhere so you can check later if it is stuck.

   Let's say we want to kill the process:

   Set procedure to ProcesssFunctions.cc
   p = new ProcessFunctions()
   hand = p.ProcessHandle(pid)  //Obtain the Handle of the process you want to kill
   exitcode = 0  //set exit code to what you want to record in system log
   killstatus = p.KillProcess(hand,exitcode)  //Kill it.  

   Gather information about a process:

   set procedure to ProcessFunctions.cc
   p = new ProcessFunctions()
   pTable = p.QueryProcess(".","4820")  //Computer Name or '.' for default and Process Id (Omit for all Processes running)
   //pTable = p.QueryProcess(".",Null)  //This will return all processes
   release object p
   p = null
   clear
   for i = 1 to pTable.size/4
   ?pTable[i,1]+" "+pTable[i,2]+" "+pTable[i,3]+" = "+pTable[i,4]
   Next

   This function will return an Array with four columns.

   1.	Record Number
   2.	Field Number
   3.	Field Name
   4.	Field Data

   For a full description of the fields and data refer to:

   http://msdn.microsoft.com/en-us/library/aa394372(v=vs.85).aspx

   Example of one Process Id:

   1 1 Caption = OUTLOOK.EXE
   1 2 CommandLine = "C:\Program Files (x86)\Microsoft Office\OFFICE11\OUTLOOK.EXE" /recycle
   1 3 CreationClassName = Win32_Process
   1 4 CreationDate = 20101215142942.786940-300
   1 5 CSCreationClassName = Win32_ComputerSystem
   1 6 CSName = VANCE001
   1 7 Description = OUTLOOK.EXE
   1 8 ExecutablePath = C:\Program Files (x86)\Microsoft Office\OFFICE11\OUTLOOK.EXE
   1 9 ExecutionState =
   1 10 Handle = 4820
   1 11 HandleCount = 999
   1 12 InstallDate =
   1 13 KernelModeTime = 43056276
   1 14 MaximumWorkingSetSize = 1380
   1 15 MinimumWorkingSetSize = 200
   1 16 Name = OUTLOOK.EXE
   1 17 OSCreationClassName = Win32_OperatingSystem
   1 18 OSName = Microsoft Windows 7 Ultimate |C:\Windows|\Device\Harddisk0\Partition2
   1 19 OtherOperationCount = 31777
   1 20 OtherTransferCount = 1461770
   1 21 PageFaults = 33233
   1 22 PageFileUsage = 38084
   1 23 ParentProcessId = 3988
   1 24 PeakPageFileUsage = 43552
   1 25 PeakVirtualSize = 392814592
   1 26 PeakWorkingSetSize = 79300
   1 27 Priority = 8
   1 28 PrivatePageCount = 38998016
   1 29 ProcessId = 4820
   1 30 QuotaNonPagedPoolUsage = 89
   1 31 QuotaPagedPoolUsage = 586
   1 32 QuotaPeakNonPagedPoolUsage = 101
   1 33 QuotaPeakPagedPoolUsage = 609
   1 34 ReadOperationCount = 3739
   1 35 ReadTransferCount = 21946059
   1 36 SessionId = 1
   1 37 Status =
   1 38 TerminationDate =
   1 39 ThreadCount = 26
   1 40 UserModeTime = 48984314
   1 41 VirtualSize = 362721280
   1 42 WindowsVersion = 6.1.7600
   1 43 WorkingSetSize = 80822272
   1 44 WriteOperationCount = 81373
   1 45 WriteTransferCount = 52618665

*/

Class ProcessFunctions

   Function ProcessID()

      if type( "GetCurrentProcessId" ) # "FP"
         extern CINT GetCurrentProcessId( CVOID ) kernel32 
      endif

   Return GetCurrentProcessId()



   Function ProcessHandle(PID)

      if type( "OpenProcess" ) # "FP"
         extern CHANDLE OpenProcess( CULONG, CLOGICAL, CINT ) kernel32 
      endif

   Return OpenProcess(0x0001,False,PID)

   Function KillProcess(HAND,EXITCODE)

      if type( "TerminateProcess" ) # "FP"
         extern CINT TerminateProcess( CHANDLE, CINT ) kernel32 
      endif

   Return TerminateProcess(HAND,EXITCODE)

   Function QueryProcess(Computer,Pid)

   	local CRLF, MSObject, VBCode, FArray, RArray, RecordString, p
	   //Windows Management Instrumentation
      CRLF = chr(13)+chr(10)
      MSObject = new OleAutoClient( "MSScriptControl.ScriptControl" )
      MSObject.language = "VBScript"
      VBCode = 'function DoQuery()'+CRLF
      VBCode += 'f = ""'+CRLF
      VBCode += 'Set objWMIService = GetObject("winmgmts:{impersonationLevel=impersonate}!\\'+computer+'\root\cimv2")'+CRLF
      //   VBCode += 'Set objWMIService = GetObject("winmgmts:\\'+computer+'\root\cimv2")'+CRLF
      VBCode += 'Set colItems = objWMIService.ExecQuery("Select * From Win32_Process")'+CRLF
      VBCode += 'For Each objItem in colItems'+CRLF
      if empty(pid) = false
         VBCode += 'If objItem.ProcessID = '+pid+' Then'+CRLF
      endif
      VBCode += 'f = f + objItem.Caption'+CRLF
      VBCode += 'f = f + "^" & objItem.CommandLine'+CRLF
      VBCode += 'f = f + "^" & objItem.CreationClassName'+CRLF
      VBCode += 'f = f + "^" & objItem.CreationDate'+CRLF
      VBCode += 'f = f + "^" & objItem.CSCreationClassName'+CRLF
      VBCode += 'f = f + "^" & objItem.CSName'+CRLF
      VBCode += 'f = f + "^" & objItem.Description'+CRLF
      VBCode += 'f = f + "^" & objItem.ExecutablePath'+CRLF
      VBCode += 'f = f + "^" & objItem.ExecutionState'+CRLF
      VBCode += 'f = f + "^" & objItem.Handle'+CRLF
      VBCode += 'f = f + "^" & objItem.HandleCount'+CRLF
      VBCode += 'f = f + "^" & objItem.InstallDate'+CRLF
      VBCode += 'f = f + "^" & objItem.KernelModeTime'+CRLF
      VBCode += 'f = f + "^" & objItem.MaximumWorkingSetSize'+CRLF
      VBCode += 'f = f + "^" & objItem.MinimumWorkingSetSize'+CRLF
      VBCode += 'f = f + "^" & objItem.Name'+CRLF
      VBCode += 'f = f + "^" & objItem.OSCreationClassName'+CRLF
      VBCode += 'f = f + "^" & objItem.OSName'+CRLF
      VBCode += 'f = f + "^" & objItem.OtherOperationCount'+CRLF
      VBCode += 'f = f + "^" & objItem.OtherTransferCount'+CRLF
      VBCode += 'f = f + "^" & objItem.PageFaults'+CRLF
      VBCode += 'f = f + "^" & objItem.PageFileUsage'+CRLF
      VBCode += 'f = f + "^" & objItem.ParentProcessId'+CRLF   
      VBCode += 'f = f + "^" & objItem.PeakPageFileUsage'+CRLF
      VBCode += 'f = f + "^" & objItem.PeakVirtualSize'+CRLF
      VBCode += 'f = f + "^" & objItem.PeakWorkingSetSize'+CRLF
      VBCode += 'f = f + "^" & objItem.Priority'+CRLF
      VBCode += 'f = f + "^" & objItem.PrivatePageCount'+CRLF
      VBCode += 'f = f + "^" & objItem.ProcessId'+CRLF
      VBCode += 'f = f + "^" & objItem.QuotaNonPagedPoolUsage'+CRLF
      VBCode += 'f = f + "^" & objItem.QuotaPagedPoolUsage'+CRLF
      VBCode += 'f = f + "^" & objItem.QuotaPeakNonPagedPoolUsage'+CRLF
      VBCode += 'f = f + "^" & objItem.QuotaPeakPagedPoolUsage'+CRLF
      VBCode += 'f = f + "^" & objItem.ReadOperationCount'+CRLF
      VBCode += 'f = f + "^" & objItem.ReadTransferCount'+CRLF
      VBCode += 'f = f + "^" & objItem.SessionId'+CRLF
      VBCode += 'f = f + "^" & objItem.Status'+CRLF
      VBCode += 'f = f + "^" & objItem.TerminationDate'+CRLF
      VBCode += 'f = f + "^" & objItem.ThreadCount'+CRLF
      VBCode += 'f = f + "^" & objItem.UserModeTime'+CRLF
      VBCode += 'f = f + "^" & objItem.VirtualSize'+CRLF
      VBCode += 'f = f + "^" & objItem.WindowsVersion'+CRLF
      VBCode += 'f = f + "^" & objItem.WorkingSetSize'+CRLF
      VBCode += 'f = f + "^" & objItem.WriteOperationCount'+CRLF
      VBCode += 'f = f + "^" & objItem.WriteTransferCount'+CRLF
      VBCode += 'f = f + "^^"'+CRLF
      if empty(pid) = false
         VBCode += 'End If'+CRLF
      endif
      VBCode += 'Next'+CRLF
      VBCode += 'DoQuery = f'+CRLF
      VBCode += 'End Function'
      MSObject.AddCode( VBCode )
      RecordString = MSObject.run( "DoQuery" )
      release object MSObject
      MSObject = null

	   FArray = new Array()

      Farray.add("Caption")
      Farray.add("CommandLine")
      Farray.add("CreationClassName")
      Farray.add("CreationDate")
      Farray.add("CSCreationClassName")
      Farray.add("CSName")
      Farray.add("Description")
      Farray.add("ExecutablePath")
      Farray.add("ExecutionState")
      Farray.add("Handle")
      Farray.add("HandleCount")
      Farray.add("InstallDate")
      Farray.add("KernelModeTime")
      Farray.add("MaximumWorkingSetSize")
      Farray.add("MinimumWorkingSetSize")
      Farray.add("Name")
      Farray.add("OSCreationClassName")
      Farray.add("OSName")
      Farray.add("OtherOperationCount")
      Farray.add("OtherTransferCount")
      Farray.add("PageFaults")
      Farray.add("PageFileUsage")
      Farray.add("ParentProcessId")
      Farray.add("PeakPageFileUsage")
      Farray.add("PeakVirtualSize")
      Farray.add("PeakWorkingSetSize")
      Farray.add("Priority")
      Farray.add("PrivatePageCount")
      Farray.add("ProcessId")
      Farray.add("QuotaNonPagedPoolUsage")
      Farray.add("QuotaPagedPoolUsage")
      Farray.add("QuotaPeakNonPagedPoolUsage")
      Farray.add("QuotaPeakPagedPoolUsage")
      Farray.add("ReadOperationCount")
      Farray.add("ReadTransferCount")
      Farray.add("SessionId")
      Farray.add("Status")
      Farray.add("TerminationDate")
      Farray.add("ThreadCount")
      Farray.add("UserModeTime")
      Farray.add("VirtualSize")
      Farray.add("WindowsVersion")
      Farray.add("WorkingSetSize")
      Farray.add("WriteOperationCount")
      Farray.add("WriteTransferCount")

      RArray = new Array()

      fc = 1
      rc = 1
      rrc = 1

      do while at("^",RecordString) > 0

         p = at("^",RecordString)

         if p > 0

            Rarray.resize(rc,4)

            RArray[rc,1] = rrc
            RArray[rc,2] = fc
            RArray[rc,3] = FArray[fc]
            RArray[rc,4] = substr(RecordString,1,p - 1)

            fc++

            rc++

            if fc > fArray.size

               fc = 1
               rrc++

               RecordString = substr(RecordString,p + 2)

            else

	            RecordString = substr(RecordString,p + 1)

            endif
         endif

      enddo

   return RArray

endclass
