/*
-----------------------------------------------------------------------
Program........: TableProp.cc
Programmer.....: Christer Persson
Date...........: 2001-09-15
Version........: 2.0
Written for....: dB2k
Rev. History...: 2001-03-04 - Original version 1.00
					  2001-09-15 - ver 2.0
                 Fixed bug for numeric fields without decimals
                 Fixed bug when writing timestamp data
                 Added fieldsByName array.
                 Added noDuplicates property.
                 Added RemoveDuplicates() method.
                 Added TrimNull() method.
                 Added fieldNumber property to the FieldDef class
                 Added support for Referential integrity
                 Added RefIntegProps array
                 Added RefIntegProp class

Usage..........: Reads and sets table field properties
Acknowledgement: The methods ReadDouble, WriteDouble, ReadLong, 
                 ReadULong, WriteLong and WriteULong partly originates
                 from BFile.cc by Jim Sare, and have been modified
                 for the DBF7 file structure.
                 There are also modifications of code from Structs.cc by
                 Vic McClung and Ken Chan.
Example........:

local f
set procedure to Tableprop.cc additive
f=new TableProp()
f.filename=cFileName
f.readHeader()
** Set properties here **

oProp=new StandardProp()
oProp.fieldNumber=2			// Field number
// Number of property: 1=Required, 2=Min, 3=Max, 4=Default
oProp.propertyNumber=3
oProp.value=950            
f.StandardProps.add(oProp)

oProp=new CustomProp()
oProp.fieldnumber=1
oProp.name='FONTBOLD'
oProp.value='true'
f.CustomProps.add(oProp)

f.writeHeader()
return
-----------------------------------------------------------------------
*/
class TableProp OF File CUSTOM

PROTECT SPDStart, CPDStart, RIStart, DataStart, ;
FPSSize, b0, b10, b64, FDTerminator, headerSize, endOfStructure, ;
CheckArray, readCPD, readCustomValues, readFDA, readFPS, readSPD, ;
readStandardValues, readTheHeader, writeCPD, writeCustomValues, ;
writeFDA, writeFPS, writeSPD, writeStandardValues, writeTheHeader, ;
readRID, readRIValues

	this.filename=''				// Name of file to edit
   this.completed=false			// true if writing to file is completed
   this.lDriver=''				// Name of language driver
   this.fields=new array()		// List of fields
   this.fieldCount=0				// Number of fields
   this.fieldsByName= new AssocArray()	// List of fields referenced by name
   this.SProperties=0			// Number of Standard properties
   this.CProperties=0			// Number of Custom properties
   this.RIProps=0					// Number of RI properties
   this.StandardProps=new array()	// List of standard properties
   this.CustomProps=new array()		// List of custom properties
   this.RefIntegProps=new array()			// List of RI properties
   this.recordCount=0			// Number of records
   this.headerRead=false		// true if header has been read
   this.errorMsg=''				// Message in case of error
   this.noDuplicates=true		// If true, only one generation of properties will be written

   // Protected properties
   this.headerSize=0				// Size of header in bytes
   this.SPDStart=0				// Start of Standard Property Descriptor Array
   this.CPDStart=0				// Start of Custom Property Descriptor Array
   this.RIStart=0					// Start of RI Property Descriptor Array
   this.DataStart=0				// Start of data
   this.FPSSize=0					// Size of Field Properties Structure
   this.b0=''						// Data starting at byte 0, not to be changed
   this.b10=''						// Data starting at byte 10, not to be changed
   this.b64=''						// Data starting at byte 64, not to be changed
   this.FDTerminator=0			// Position of Field Descriptor Terminator
   this.endOfStructure=0		// Position of end of Field Properties Structure
   this.readError=false			// true if a read-error occurs

function readHeader
	try
   	this.readTheHeader()
   catch (Exception e)
   	this.close()
      this.errorMsg=e.message+' on line '+e.lineno
   endtry
   return

function readTheHeader
	this.open(this.filename)
   this.b0=this.read(1)
   // Check that the file is level 7
   if BITSET(ASC(this.b0), 0) or BITSET(ASC(this.b0), 1) ;
   or not BITSET(ASC(this.b0), 2)
      this.errorMsg='Not a dBASE level 7 file.'
	   this.close()
      return
   endif
   this.read(3)
   this.recordCount=this.read32()	// Number of records
   this.headerSize=this.read16()		// Number of bytes in header
   this.b10=this.read(22)			// read data up to Language driver name
   this.lDriver=this.read(32)		// Language driver name
   this.b64=this.read(4)
   this.readFDA()						// Read Field Descriptor Array
   // Is there a Field Properties array?
   if this.position<=this.headerSize-16
   	this.readFPS()					  	// Read Field Properties Structure
      this.readSPD()						// Read Standard Property Descriptor Array
      this.readCPD()						// Read Custom Property Descriptor Array
      this.readRID()						// Read Ref Integrity Prop Descr Array
      if this.readError
	      this.close()
      	return
      endif
      this.readStandardValues()
      this.readCustomValues()
      this.readRIValues()
   endif
   this.headerRead=true
   this.close()
	return

function CheckArray
	// Check that all standard properties have valid field number and property number
   // Otherwise delete property
   local i
   // Some properties must be integers
   for i=1 to this.StandardProps.size
   	this.StandardProps[i].fieldnumber=INT(this.StandardProps[i].fieldnumber)
   	this.StandardProps[i].propertyNumber=INT(this.StandardProps[i].propertyNumber)
   endfor
   for i=1 to this.CustomProps.size
   	this.CustomProps[i].fieldnumber=INT(this.CustomProps[i].fieldnumber)
   	this.CustomProps[i].fieldType=INT(this.CustomProps[i].fieldType)
   endfor
   for i=this.StandardProps.size to 1 step -1
   	do case
      	// Check that fieldnumber is within range
      	case this.StandardProps[i].fieldnumber<1 or ;
	      this.StandardProps[i].fieldnumber>this.fields.size
            this.StandardProps.delete(i)
            this.StandardProps.size--
            this.Sproperties--
         // Fieldtypes Character, Numeric, Date, Float, Long, Timestamp, Double
      	case this.fields[this.StandardProps[i].fieldnumber].fieldtype='C' or ;
         this.fields[this.StandardProps[i].fieldnumber].fieldtype='N' or ;
         this.fields[this.StandardProps[i].fieldnumber].fieldtype='D' or ;
         this.fields[this.StandardProps[i].fieldnumber].fieldtype='F' or ;
         this.fields[this.StandardProps[i].fieldnumber].fieldtype='I' or ;
         this.fields[this.StandardProps[i].fieldnumber].fieldtype='@' or ;
         this.fields[this.StandardProps[i].fieldnumber].fieldtype='O'
				// Property numbers 1-4 allowed
            if this.StandardProps[i].propertyNumber<1 or ;
            this.StandardProps[i].propertyNumber>4
               this.StandardProps.delete(i)
               this.StandardProps.size--
               this.Sproperties--
            endif
         // Fieldtype Logical
      	case this.fields[this.StandardProps[i].fieldnumber].fieldtype='L'
				// Property numbers 1 and 4 allowed
            if this.StandardProps[i].propertyNumber<>1 and ;
            this.StandardProps[i].propertyNumber<>4
               this.StandardProps.delete(i)
               this.StandardProps.size--
               this.Sproperties--
            endif
         // Fieldtypes Memo, OLE, Binary
      	case this.fields[this.StandardProps[i].fieldnumber].fieldtype='M' or ;
         this.fields[this.StandardProps[i].fieldnumber].fieldtype='G' or ;
         this.fields[this.StandardProps[i].fieldnumber].fieldtype='B'
				// Property number 1 allowed
            if this.StandardProps[i].propertyNumber<>1
               this.StandardProps.delete(i)
               this.StandardProps.size--
               this.Sproperties--
            endif
         otherwise
            this.StandardProps.delete(i)
            this.StandardProps.size--
            this.Sproperties--
      endcase
   endfor
	// Check that all custom properties have valid field number and name
   // Otherwise delete property
   for i=this.CustomProps.size to 1 step -1
   	if this.CustomProps[i].fieldnumber<1 or ;
      this.CustomProps[i].fieldnumber>this.fields.size or ;
      this.CustomProps[i].name==''
      	this.CustomProps.delete(i)
         this.CustomProps.size--
         this.Cproperties--
      endif
   endfor
   return

function RemoveDuplicates
	// Allow only one generation of properties
   local tempStandardProps, tempCustomProps, i, oProp, j
   // Make sure also newly added properties has a generation number.
   // The last property gets the highest generation number.
   for i=1 to this.StandardProps.size
   	this.StandardProps[i].generation=i
   endfor
   for i=1 to this.CustomProps.size
   	this.CustomProps[i].generation=i
   endfor

   tempStandardProps=new array()		// array to collect standard properties
   tempCustomProps=new array()		// array to collect custom properties
   // Go through standard properties
   for i=1 to this.StandardProps.size
   	oProp=this.StandardProps[i]
     	// Does this property already exist for this field?
      for j=1 to tempStandardProps.size
      	if oProp.fieldnumber=tempStandardProps[j].fieldnumber and ;
      	oProp.propertyNumber=tempStandardProps[j].propertyNumber
         	exit
         endif
      endfor
      // If no matching property is found, then j>tempStandardProps.size
      if j>tempStandardProps.size
      	tempStandardProps.add(oProp)		// Add property
      else
      	// If generation is greater, then replace
      	if oProp.generation>tempStandardProps[j].generation
         	tempStandardProps[j]=oProp
         endif
      endif
   endfor
   // Go through custom properties
   for i=1 to this.CustomProps.size
   	oProp=this.CustomProps[i]
     	// Does this property already exist for this field?
      for j=1 to tempCustomProps.size
      	if oProp.fieldnumber=tempCustomProps[j].fieldnumber and ;
      	UPPER(oProp.name)=UPPER(tempCustomProps[j].name)
         	exit
         endif
      endfor
      // If no matching property is found, then j>tempCustomProps.size
      if j>tempCustomProps.size
      	tempCustomProps.add(oProp)		// Add property
      else
      	// If generation is greater, then replace
      	if oProp.generation>tempCustomProps[j].generation
         	tempCustomProps[j]=oProp
         endif
      endif
   endfor
   // Replace this.StandardProps and this.CustomProps with temporary arrays
   this.StandardProps=tempStandardProps
   this.CustomProps=tempCustomProps
   this.SProperties=this.StandardProps.size
   this.CProperties=this.CustomProps.size
	return

function writeHeader
	try
   	this.writeTheHeader()
   catch (Exception e)
   	this.close()
      this.errorMsg=e.message+' on line '+e.lineno
   endtry
   return

function writeTheHeader
	local cTempFile, i, nPad
	if not this.headerRead
      this.errorMsg='Not completed. Header not read.'
   	return this.completed
   endif
   if this.recordCount>0		// Is number of records > 0 ?
      this.errorMsg='Not completed. Can only operate on files without data records.'
   	return this.completed
   endif
   this.CheckArray()			// Check that property arrays are OK
   if this.noDuplicates
   	this.RemoveDuplicates()
   endif
   cTempFile=FUNIQUE('temp????.dbf')
   this.create(cTempFile)
   this.write(this.b0)				// version etc.
   // Date of last update, YYMMDD
	this.write(CHR(VAL(SUBSTR(STR(YEAR(DATE()),4),3,2))))
   this.write(CHR(MONTH(DATE())))
   this.write(CHR(DAY(DATE())))
	// Number of records, always 0
   this.Write32(0)
   // Number of bytes in header, to be changed later
   this.Write16(0)
   this.write(this.b10)			// byte 10-31
   // Language driver name, byte 32-63
   this.write(LEFT(this.lDriver, 32)+REPLICATE(CHR(0), 32-LEN(this.lDriver)))
   this.write(this.b64)			// byte 64-67
   // Write Field Descriptor Array
  	this.writeFDA()
   this.write(CHR(13))			// Field Descriptor terminator
   this.FDTerminator=this.position	// Position of Field Descriptor terminator+1
   // Field Properties Structure
   if this.StandardProps.size>0 or this.CustomProps.size>0 ;
   or this.RefIntegProps.size>0
   	this.writeFPS()
      this.writeSPD()
      this.writeCPD()
      this.writeRID()
      this.DataStart=this.position-this.FDTerminator
      this.writeStandardValues()
      this.writeCustomValues()
      this.writeRIValues()
   endif
   this.endOfStructure=this.position
   // Pad with 0x00 to the nearest 0x200
   nPad=MOD(this.endOfStructure-this.FDTerminator+1, 512)
   nPad=512-nPad
   this.write(REPLICATE(CHR(0), nPad))
   this.write(CHR(26))			// EOF
   this.headerSize=this.position
   // Write headerSize
   this.seek(8)
   this.write16(this.headerSize)

   if this.StandardProps.size>0 or this.CustomProps.size>0 ;
   or this.RefIntegProps.size>0
      // Start of Standard Property Descriptor Array
      this.seek(this.FDTerminator+2)
      this.write16(this.SPDstart)
      // Start of Custom Property Descriptor Array
      this.seek(2, 1)
      this.write16(this.CPDstart)
      // Start of Referential Integrity Property Descriptor Array
      this.seek(2, 1)
      this.write16(this.RIStart)
      // Start of data
      this.write16(this.DataStart)
      // Size of Field Properties Structure
      this.write16(this.endOfStructure-this.FDTerminator)

      // Write start of data for each Standard Property
      this.seek(this.SPDstart+this.FDTerminator)
      for i=1 to this.StandardProps.size
         this.seek(11, 1)
         this.write16(this.StandardProps[i].dataStart)
         this.seek(2, 1)
      endfor

      // Write start of name and data for each Custom Property
      this.seek(this.CPDstart+this.FDTerminator)
      for i=1 to this.CustomProps.size
         this.seek(6, 1)
         this.write16(this.CustomProps[i].nameStart)
         this.seek(2, 1)
         this.write16(this.CustomProps[i].dataStart)
         this.seek(2, 1)
      endfor

      // Write start of rule name, foreign table name, local table tag name
      // and foreign table tag name
      this.seek(this.RIStart+this.FDTerminator)
      for i=1 to this.RefIntegProps.size
         this.seek(3, 1)
         this.write16(this.RefIntegProps[i].RIRuleNameStart)
         this.seek(2, 1)
         this.write16(this.RefIntegProps[i].ForeignNameStart)
         this.seek(5, 1)
         this.write16(this.RefIntegProps[i].localTagNameStart)
         this.seek(2, 1)
         this.write16(this.RefIntegProps[i].foreignTagNameStart)
         this.seek(2, 1)
      endfor
   endif

   this.close()
   this.headerRead=false
   delete file (this.filename)
   rename (cTempFile) to (this.filename)
   this.completed=true
	return this.completed

function writeRID
	// Write RI Property Descriptor Array
   local i
   this.RIStart=this.position-this.FDTerminator
   for i=1 to this.RefIntegProps.size
   	if this.RefIntegProps[i].master
      	this.write(CHR(7))
      else
      	this.write(CHR(8))
      endif
   	this.write16(i)			// Sequential number
      this.write16(0)			// Start of rule name, to be changed
      this.write16(LEN(this.RefIntegProps[i].RIRuleName)+1)	// Length of rule name
      this.write16(0)			// Start of Foreign table name, to be changed
      // Length of Foreign table name
      this.write16(LEN(this.RefIntegProps[i].ForeignName)+1)
      this.write(this.RefIntegProps[i].UpdateDelete)	// Update/delete behaviour
      // Number of fields in linking key
      this.write16(this.RefIntegProps[i].linkingFields)
      this.write16(0)			// Start of Local table tag name, to be changed
      // Length of Local table tag name
      this.write16(LEN(this.RefIntegProps[i].localTagName)+1)
      this.write16(0)			// Start of Foreign table tag name, to be changed
      // Length of Foreign table tag name
      this.write16(LEN(this.RefIntegProps[i].foreignTagName)+1)
   endfor
	return

function writeRIValues
	local i
   for i=1 to this.RefIntegProps.size
     	this.RefIntegProps[i].RIRuleNameStart=this.position-this.FDTerminator
   	this.write(this.RefIntegProps[i].RIRuleName)
      this.write(CHR(0))
     	this.RefIntegProps[i].ForeignNameStart=this.position-this.FDTerminator
      this.write(this.RefIntegProps[i].ForeignName)
      this.write(CHR(0))
     	this.RefIntegProps[i].localTagNameStart=this.position-this.FDTerminator
      this.write(this.RefIntegProps[i].localTagName)
      this.write(CHR(0))
     	this.RefIntegProps[i].foreignTagNameStart=this.position-this.FDTerminator
      this.write(this.RefIntegProps[i].foreignTagName)
      this.write(CHR(0))
   endfor
	return

function writeCustomValues
	local i
   for i=1 to this.CustomProps.size
     	this.CustomProps[i].nameStart=this.position-this.FDTerminator
   	this.write(this.CustomProps[i].name)
     	this.CustomProps[i].dataStart=this.position-this.FDTerminator
      this.write(this.CustomProps[i].value)
      this.write(CHR(0))
   endfor
	return

function writeStandardValues
	// Write Standard Property data
   local i, nInt, nDec, cPicture, nLength, cVal
   private TempVal
   for i=1 to this.StandardProps.size
   	if this.StandardProps[i].propertyNumber=1		// Property: Required
      	this.StandardProps[i].dataStart=0
      else
      	this.StandardProps[i].dataStart=this.position-this.FDTerminator
         TempVal=this.StandardProps[i].value
         do case

         	// Undefined or Character
         	case this.StandardProps[i].fieldType=0 or ;
            this.StandardProps[i].fieldType=1

               nLength=this.fields[this.StandardProps[i].fieldNumber].length
            	if TYPE('TempVal')=='C'
                  // Set datalength the same as length of field
                  this.StandardProps[i].datalength=nLength
                  // Ensure value is not too long
                  this.StandardProps[i].value=;
                  LEFT(this.StandardProps[i].value, nLength)
                  // Pad with CHR(0) if too short
                  this.StandardProps[i].value=this.StandardProps[i].value+;
                  REPLICATE(CHR(0), nLength-LEN(this.StandardProps[i].value))
                  // Write value
                  this.write(this.StandardProps[i].value)
               else
               	this.write(REPLICATE(CHR(0),nLength))
               endif

            // Numeric or Float
         	case this.StandardProps[i].fieldType=2 or ;
            this.StandardProps[i].fieldType=6

               nLength=this.fields[this.StandardProps[i].fieldNumber].length
            	if TYPE('TempVal')=='N'
                  // Number of decimals
                  nDec=this.fields[this.StandardProps[i].fieldNumber].decimals
                  // Number of integers
                  nInt=nLength-nDec-1
                  // Create picture
                  if nDec=0
	                  cPicture=REPLICATE('9', nLength)
                     cVal=TRANSFORM(this.StandardProps[i].value, cPicture)		// added ver 1.01
                  else
	                  cPicture=REPLICATE('9', nInt)+'.'+REPLICATE('9', nDec)
                     cVal=TRANSFORM(this.StandardProps[i].value, cPicture)
                     if SET('POINT')<>'.'
                        if AT(SET('POINT'), cVal)>0
                        	cVal=STUFF(cVal, AT(SET('POINT'), cVal), 1, '.')
                        endif
                     endif
                  endif
                  // Write value
                  this.write(cVal)
               else
               	this.write(REPLICATE(CHR(0),nLength))
               endif

            // Logical
         	case this.StandardProps[i].fieldType=4

            	if TYPE('TempVal')=='L'
               	if this.StandardProps[i].value
                  	this.write('T')
                  else
                  	this.write('F')
                  endif
               else
               	this.write(CHR(0))
               endif

            // Date
         	case this.StandardProps[i].fieldType=5

            	if TYPE('TempVal')=='D'
                  this.write(TRANSFORM(YEAR(TempVal), '@L 9999'))
                  this.write(TRANSFORM(MONTH(TempVal), '@L 99'))
                  this.write(TRANSFORM(DAY(TempVal), '@L 99'))
               else
               	this.write(REPLICATE(CHR(0), 8))
               endif

            // Long
         	case this.StandardProps[i].fieldType=11

            	if TYPE('TempVal')=='N'
               	this.writeLong(TempVal)
               else
               	this.write(REPLICATE(CHR(0), 4))
               endif

            // Timestamp
         	case this.StandardProps[i].fieldType=12

            	if TYPE('TempVal')=='DT' or TYPE('TempVal')=='O'
               	this.write(this.DT2byte(TempVal))
               else
               	this.write(REPLICATE(CHR(0), 8))
               endif

            // Double
         	case this.StandardProps[i].fieldType=13

            	if TYPE('TempVal')=='N'
               	this.writeDouble(TempVal)
               else
               	this.write(REPLICATE(CHR(0), 8))
               endif

            otherwise
		      	this.StandardProps[i].dataStart=0
         endcase
      endif
   endfor
	return

function writeFPS
   // Write field Properties Structure
   this.write16(this.StandardProps.size)			// Number of Standard properties
   this.write16(0)	// Start of Standard Property Descriptor Array, to be changed
   this.write16(this.CustomProps.size)			// Number of Custom properties
   this.write16(0)	// Start of Custom Property Descriptor Array, to be changed
   this.write16(this.RefIntegProps.size)			// Number of RI properties
   this.write16(0)		// Start of RI Property Descriptor Array, to be changed
   this.write16(0)					// Start of data, to be changed later
   this.write16(0)					// Size of structure, to be changed later
	return

function writeSPD
	// Write Standard Property Descriptor Array
   local i
   this.SPDStart=this.position-this.FDTerminator
   for i=1 to this.StandardProps.size
   	this.write16(i)			// Generational number
      this.write16(this.StandardProps[i].fieldNumber)
      this.write(CHR(this.StandardProps[i].propertyNumber))
      this.StandardProps[i].fieldType=;
      this.FieldTypeNo(this.fields[this.StandardProps[i].fieldNumber].fieldtype)
      this.write(CHR(this.StandardProps[i].fieldType))
      this.StandardProps[i].b06=;
      IIF(this.StandardProps[i].propertyNumber=6, CHR(0x00), CHR(0x02))
      this.write(this.StandardProps[i].b06)
      this.write(REPLICATE(CHR(0), 4))
      this.write16(0)			// Start of data, to be changed
		// Width of field
      this.write16(this.fields[this.StandardProps[i].fieldNumber].length)
   endfor
	return

function writeCPD
	// Write Custom Property Descriptor Array
   local i
   this.CPDStart=this.position-this.FDTerminator
   for i=1 to this.CustomProps.size
   	this.write16(i)			// Generational number
      this.write16(this.CustomProps[i].fieldNumber)
      this.write(CHR(this.CustomProps[i].fieldType))
      this.write(CHR(0))
      this.write16(0)			// Start of name, to be changed
      this.write16(LEN(this.CustomProps[i].name))	// Length of name
      this.write16(0)			// Start of data, to be changed
     	// Length of data + null terminator
      this.write16(LEN(this.CustomProps[i].value)+1)
   endfor
	return

function writeFDA
   // Write Field Descriptor Array
	local i
   for i=1 to this.fields.size
      this.write(this.fields[i].fieldname)
      this.write(this.fields[i].fieldtype)
      this.write(CHR(this.fields[i].length))
      this.write(CHR(this.fields[i].decimals))
      this.write(this.fields[i].rest)
   endfor
	return

function readRID
	// Read Referential Integrity Property Descriptor array
   local i, oProp
   for i=1 to this.RIProps
   	oProp=new RefIntegProp()
      oProp.MasterDep=this.read(1)			// The code for Master/Dependent
      do case
      	case ASC(oProp.MasterDep)=7
         	oProp.Master=true
      	case ASC(oProp.MasterDep)=8
         	oProp.Master=false
			otherwise
         	this.readError=true
            return
      endcase
      oProp.SeqNum=this.read16()				// Sequential number
      oProp.RIRuleNameStart=this.read16()		// Start position of RI rule name
      oProp.RIRuleNameLength=this.read16()		// Length of RI rule name
      // Start position of the name of the foreign table
      oProp.ForeignNameStart=this.read16()
      // Length of the name of the foreign table
      oProp.ForeignNameLength=this.read16()
      oProp.UpdateDelete=this.read(1)

      *      * Update   * Delete   *
      *------*----------*----------*
      * 0x00 * Restrict * Restrict *
      * 0x01 * Restrict * Cascade  *
      * 0x10 * Cascade  * Restrict *
      * 0x11 * Cascade  * Cascade  *

      oProp.linkingFields=this.read16()		// Seems to always be 0
      oProp.localTagNameStart=this.read16()		// Start of local table tag name
      oProp.localTagNameLength=this.read16()		// Length of local table tag name 
      oProp.foreignTagNameStart=this.read16()	// Start of foreign table tag name
      oProp.foreignTagNameLength=this.read16()	// Length of foreign table tag name
   	this.RefIntegProps.add(oProp)
   endfor
   return

function readRIValues
	// Read names and data for RI properties
   local i
   for i=1 to this.RIProps
   	this.seek(this.FDTerminator+this.RefIntegProps[i].RIRuleNameStart)
   	this.RefIntegProps[i].RIRuleName=;
      this.read(this.RefIntegProps[i].RIRuleNameLength-1)

   	this.seek(this.FDTerminator+this.RefIntegProps[i].ForeignNameStart)
   	this.RefIntegProps[i].ForeignName=;
      this.read(this.RefIntegProps[i].ForeignNameLength-1)

   	this.seek(this.FDTerminator+this.RefIntegProps[i].localTagNameStart)
   	this.RefIntegProps[i].localTagName=;
      this.read(this.RefIntegProps[i].localTagNameLength-1)

   	this.seek(this.FDTerminator+this.RefIntegProps[i].foreignTagNameStart)
   	this.RefIntegProps[i].foreignTagName=;
      this.read(this.RefIntegProps[i].foreignTagNameLength-1)
   endfor
	return

function readCPD
	// Read Custom Property Descriptor array
   local i, oProp
   for i=1 to this.CProperties
   	oProp=new CustomProp()
      oProp.generation=this.read16()			// Generation
      oProp.fieldNumber=this.read16()		// Field number
      oProp.fieldType=ASC(this.read(1))		// Field type
      oProp.b05=this.read(1)					// Data at byte 5, not to be changed
      oProp.nameStart=this.read16()			// Start position of name
      oProp.nameLength=this.read16()			// Length of name
      oProp.dataStart=this.read16()			// Start position of data
      oProp.dataLength=this.read16()			// Length of data
   	this.CustomProps.add(oProp)
   endfor
   return

function readCustomValues
	// Read names and data for Custom properties
   local i
   for i=1 to this.CProperties
   	this.seek(this.FDTerminator+this.CustomProps[i].nameStart)
   	this.CustomProps[i].name=this.read(this.CustomProps[i].nameLength)
   	this.seek(this.FDTerminator+this.CustomProps[i].dataStart)
   	this.CustomProps[i].value=this.read(this.CustomProps[i].dataLength-1)
   endfor
	return

function readSPD
	// Read Standard Property Descriptor array
   local i, oProp
   for i=1 to this.SProperties
   	oProp=new StandardProp()
      oProp.generation=this.read16()				// Generation
      oProp.fieldNumber=this.read16()			// Field number
      oProp.propertyNumber=ASC(this.read(1))	// number of property
      oProp.fieldType=ASC(this.read(1))			// Field type
      oProp.b06=this.read(1)						// Data at byte 6, not to be changed
      this.read(4)
      oProp.dataStart=this.read16()				// Start position of data
      oProp.dataLength=this.read16()				// Length of data
   	this.StandardProps.add(oProp)
   endfor
   return

function readStandardValues
	// Read data for Standard properties
   local i, cTemp
   for i=1 to this.SProperties
   	if this.StandardProps[i].propertyNumber=1		// Property 1: Required
      	this.StandardProps[i].value=''		  // No data for Required property
      else
         this.seek(this.FDTerminator+this.StandardProps[i].datastart)
         do case

         	// Undefined or Character
         	case this.StandardProps[i].fieldType=0 or ;
            this.StandardProps[i].fieldType=1

		         this.StandardProps[i].value=;
               this.read(this.StandardProps[i].dataLength)

            // Numeric or Float
         	case this.StandardProps[i].fieldType=2 or ;
            this.StandardProps[i].fieldType=6

		         this.StandardProps[i].value=;
               VAL(this.read(this.StandardProps[i].dataLength))
            	

            // Logical
         	case this.StandardProps[i].fieldType=4

		         cTemp=this.read(1)
               do case
               	case UPPER(cTemp)='T'
                  	this.StandardProps[i].value=true
               	case UPPER(cTemp)='F'
                  	this.StandardProps[i].value=false
               	otherwise
                  	this.StandardProps[i].value=null
               endcase

            // Date
         	case this.StandardProps[i].fieldType=5

            	this.StandardProps[i].value=this.ReadDate()

            // Long
         	case this.StandardProps[i].fieldType=11

            	this.StandardProps[i].value=this.ReadLong()

            // Timestamp
         	case this.StandardProps[i].fieldType=12

            	this.StandardProps[i].value=this.Byte2DT(this.Read(8))

            // Double
         	case this.StandardProps[i].fieldType=13

            	this.StandardProps[i].value=this.ReadDouble()

            otherwise
		      	this.StandardProps[i].dataStart=0
         endcase
      endif
   endfor
	return

function readFPS
	// Read Field Properties Structure
   this.SProperties=this.read16()
   this.SPDStart=this.read16()
   this.CProperties=this.read16()
   this.CPDStart=this.read16()
   this.RIProps=this.read16()
   this.RIStart=this.read16()
   this.DataStart=this.read16()
   this.FPSSize=this.read16()
   return

function readFDA
   // Read Field Descriptor Array
	local oField, cTerm
   do while this.position+48<this.headerSize
   	cTerm=this.read(1)
      if cTerm=CHR(13)
      	exit
      endif
   	this.fieldCount++
      oField=new FieldDef()
      oField.fieldname=cTerm+this.read(31)
      oField.fieldtype=this.read(1)
      oField.length=ASC(this.read(1))
      oField.decimals=ASC(this.read(1))
      oField.rest=this.read(13)
      oField.fieldNumber=this.fields.size+1
   	this.fields.add(oField)
      this.fieldsByName[this.TrimNull(oField.fieldname)]=oField
   enddo
   this.FDTerminator=this.position		// Position of Field Descriptor Terminator
   return

function FieldTypeNo(c)
	// Translate field type as character to number
	c=TRIM(UPPER(LEFT(c, 1)))
   return AT(c, 'CNMLDF GB I@O+')

function Read16
	return ASC(this.read(1))+ASC(this.read(1))*256

function Read32
	return ASC(this.read(1))+ASC(this.read(1))*256+ASC(this.read(1))*256*256+;
   ASC(this.read(1))*256*256*256


function Write16(num)
	// Write 16-bit number
	this.write(CHR(MOD(num, 256)))
   this.write(CHR(INT(num/256)))
   return

function Write32(num)
	// Write 32-bit number
	local byte0, byte1, byte2, byte3
   byte3=INT(num/(256*256*256))
   num=num-(byte3*256*256*256)
   byte2=INT(num/(256*256))
   num=num-(byte2*256*256)
   byte1=INT(num/(256))
   byte0=num-(byte1*256)
   this.write(CHR(byte0))
   this.write(CHR(byte1))
   this.write(CHR(byte2))
   this.write(CHR(byte3))
   return


function ReadDouble			// Read 64-bit float
   local c, bPos, nExp, nMant, nByte, i

   c = this.Read(8)
   do case
      case c==chr(128)+replicate(chr(0), 7)
         return 0
      case c==replicate(chr(0), 8)
         return null
   endcase
   nByte=new array(8)
   for i=1 to 8
      nByte[i] = Asc(SubStr(c, i))
   endfor
   bPos = BitSet(nByte[1], 7)
   if not bPos
      for i=1 to 8
         nByte[i]=BitXOR(nByte[i], 0xFF)
      endfor
   endif
   nExp = BitLShift(BitAnd(nByte[1], 0x7F), 4)
   nExp += BitZRShift(nByte[2], 4)
   nMant = 0x10 + BitAnd(nByte[2], 0x0F)
   For i = 3 To 8
      nMant *= 256.0
      nMant +=nByte[i]
   EndFor
   nMant *= 2^(nExp - 1023 - 52)
   return IIf(bPos, nMant, -nMant)

function WriteDouble(n)		// Write 64-bit float
   local bPos, nAbs, nExp, nMant, c, nByte
   do case
      case n==null
         return this.write(REPLICATE(CHR(0), 8))
      case n==0
         return this.write(CHR(128)+REPLICATE(CHR(0), 7))
   endcase
   bPos=n>0
   nAbs = Abs(n)
   nExp = Floor(Log(nAbs) / Log(2))
   nMant = Floor(nAbs / 2^(nExp - 52))
   nExp += 1023
   c=''
   For nInd = 1 to 6
      nByte=nMant % 0x100
      if not bPos
         nByte=BitXOR(nByte, 0xFF)
      endif
      c = Chr(nByte) + c
      nMant /= 0x100
   EndFor
   nMant := BitAnd(nMant, 0x0F)
   nByte=nMant + BitLShift(BitAnd(nExp, 0x0F), 4)
   if not bPos
      nByte=BitXOR(nByte, 0xFF)
   endif
   c = Chr(nByte) + c
   nExp := BitZRShift(nExp, 4)
   nByte=nExp + 0x80
   if not bPos
      nByte=BitXOR(nByte, 0xFF)
   endif
   c = Chr(nByte) + c
   return this.Write(c)


function ReadULong			// Read 32-bit unsigned numeric
   local c
   c = this.Read(4)
   return Int(Asc(SubStr(c, 4)) +;
   BitLShift(Asc(SubStr(c, 3)),  8) +;
   BitLShift(Asc(SubStr(c, 2)), 16) +;
   BitLShift(Asc(SubStr(c, 1)), 24))

function WriteULong(n)		// Write 32-bit unsigned numeric
   return this.Write(;
   Chr(BitAnd(BitRShift(n, 24), 0xFF)) +;
   Chr(BitAnd(BitRShift(n, 16), 0xFF)) +;
   Chr(BitAnd(BitRShift(n,  8), 0xFF)) +;
   Chr(BitAnd(n, 0xFF)))

function ReadLong				// Read 32-bit signed numeric
   local n
   n = this.ReadULong()
   return n-2147483648

function WriteLong(n)		// Write 32-bit signed numeric
   return this.WriteULong(IIf(n < 0, n+2147483648, n-2147483648))

function ReadDate
	// Read a Date field
	local c, d
   c=this.read(8)
   if c==SPACE(8)
   	return {}
   endif
   d=new date()
   d.year=val(left(c, 4))
   d.month=val(substr(c, 5, 2))-1
   d.date=val(substr(c, 7))
   return ctod(left(d.toLocaleString( ), at(' ', d.toLocaleString( ))-1))

function WriteDate(d)
	// Write a Date field
   local cYear, cMonth, cDay
   if isblank(d)
   	this.write(replicate(chr(32), 8))
   else
   	cYear=TRANSFORM(YEAR(d), '@L 9999')
      cMonth=TRANSFORM(MONTH(d), '@L 99')
      cDay=TRANSFORM(DAY(d), '@L 99')
  		// Ensure length is 8, otherwise write blank date
      if LEN(cYear+cMonth+cDay)=8
	   	this.write(cYear+cMonth+cDay)
      else
	   	this.write(replicate(chr(32), 8))
      endif
   endif
   return

function Byte2DT(c1)
	// Translates an 8 byte string as stored in DBF7 table to a datetime
   local bNeg, nExp, nMant, n, nDt
   if c1==replicate(chr(0), 8)
   	return null
   endif
   c=''
   for i=8 to 1 step -1
      c=c+substr(c1, i, 1)
   endfor
   n = asc(substr(c, 8, 1))
   bNeg = bitset( n, 7 )  // Get sign from high bit of high byte
   nExp = bitlshift( bitand( n, 0x7F ), 4 )
   n = asc(substr(c, 7, 1))
   nExp += bitzrshift( n, 4 )
   // Mantissa has implied leading 1 bit
   nMant = 0x10 + bitand( n, 0x0F )
   for i = 6 to 1 step -1
      // Workaround: Must multiply by floating point number
		// to force promotion to double from integer
      nMant *= 256.0
      nMant += asc(substr(c, i, 1))
   endfor
   nMant *= 2^( nExp - 0x3ff - 52 )
   nDt=-62135686800000+iif( bNeg, -nMant, nMant )
   return new Date(nDt)

function DT2Byte(dt)
	// Translates a datetime to an 8 byte string as stored in DBF7 table
   local bPos, nAbs, nExp, nMant, c, nByte, nInd, nValue
   if dt==null
   	return replicate(chr(0), 8)
   endif
   nValue=((dt-this.dayZero())*86400000)
   bPos=nValue>0
   nAbs = Abs(nValue)
   nExp = Floor(Log(nAbs) / Log(2))
   nMant = Floor(nAbs / 2^(nExp - 52))
   nExp += 1023
   c=''
   For nInd = 1 to 6
      nByte=nMant % 0x100
      c = Chr(nByte) + c
      nMant /= 0x100
   EndFor
   nMant := BitAnd(nMant, 0x0F)
   nByte=nMant + BitLShift(BitAnd(nExp, 0x0F), 4)
   c = Chr(nByte) + c
   nExp := BitZRShift(nExp, 4)
   nByte=nExp + iif(bPos, 0, 0x80)
   c = Chr(nByte) + c
   return c

function dayZero
	//Returns the datetime 0001-01-00 BC 00:00:00
	return new Date(-62135686800000)

function TrimNull(cString, cDirection)
	// Removes CHR(0) from cString
   // If cDirection is 'r' or 'R', characters are removed from the end
   // If cDirection is 'l' or 'L', characters are removed from the beginning
   // If cDirection is missing or something else than above, characters are 
   // removed from both beginning and end
   local cString1
	if pcount()<1
   	return
   endif
	if pcount()<2
   	cDirection=''
   endif
   cString1=cString
   if UPPER(cDirection)<>'L'
   	do while asc(right(cString1, 1))=0
      	cString1=left(cString1, len(cString1)-1)
      enddo
   endif
   if UPPER(cDirection)<>'R'
   	do while asc(left(cString1, 1))=0
      	cString1=substr(cString1, 2)
      enddo
   endif
	return cString1

endclass

class FieldDef

   this.fieldname=''
   this.fieldtype=''
   this.length=0
   this.decimals=0
   this.rest=''
   this.fieldNumber=0

endclass

class CustomProp

	this.name=''				// Name of property, set by user
   this.value=null			// Value of property, set by user
	this.generation=0			// Generation
   this.fieldNumber=0		// Field number, set by user
   this.fieldType=1			// Field type, may be set by user
   this.b05=''					// Data at byte 5, not to be changed
   this.nameStart=0			// Start position of name
   this.nameLength=0			// Length of name
   this.dataStart=0			// Start position of data
   this.dataLength=0			// Length of data

endclass

class StandardProp

	this.generation=0			// Generation
   this.fieldNumber=0		// Field number
   this.propertyNumber=0	// number of property
   this.fieldType=0			// Field type
   this.b06=''					// Data at byte 6, not to be changed
   this.dataStart=0			// Start position of data
   this.dataLength=0			// Length of data
   this.value=null			// Value of property

endclass

class RefIntegProp

	this.Master=true			// true if Master (parent), false if Dependent (child)
   this.MasterDep=0x07		// The code for Master/Dependent
   this.SeqNum=0				// Sequential number
   this.RIRuleNameStart=0		// Start position of RI rule name
   this.RIRuleNameLength=0		// Length of RI rule name
   this.ForeignNameStart=0		// Start position of the name of the foreign table
   this.ForeignNameLength=0	// Length of the name of the foreign table
	this.UpdateDelete=0x00		// Update/delete behaviour

   *      * Update   * Delete   *
   *------*----------*----------*
   * 0x00 * Restrict * Restrict *
   * 0x01 * Restrict * Cascade  *
   * 0x10 * Cascade  * Restrict *
   * 0x11 * Cascade  * Cascade  *

   this.linkingFields=0				// Seems to always be 0
   this.localTagNameStart=0		// Start of local table tag name
   this.localTagNameLength=0		// Length of local table tag name - 0 terminated
   this.foreignTagNameStart=0		// Start of foreign table tag name
   this.foreignTagNameLength=0	// Length of foreign table tag name - 0 terminated
   this.RIRuleName=''				// The RI rule name
   this.ForeignName=''				// Name of foreign table
   this.localTagName=''				// Local table tag name
   this.foreignTagName=''			// Foreign table tag name

endclass
