/*
   --------------------------------------------------------
   Filename.....: fontDlg.cc
   ClassName....: fontDlg
   Purpose......: ChooseFont dialog
                  with some properties exposed.
   Programmer...: Rick Miller -  http://www.treturn.com/dbase
   Date.........: December 03, 2004
   Notes........: See Notes below.
   Written for..: dBASE dB2K 0.1 and newer.
   Rev. History.: See Revisions below.
   Dependencies.: applicationIni.cc,
                  VStruct32Ex.cc
                  or VStruct.cc,
                  or VStruct.prg,
                  and VStruct.h.
   Usage........: set procedure to fontDlg.cc additive
   HelpFile.....: None.
   Example:.....: See Example below.
   SampleFile...: fontDlg.wfm.
   --------------------------------------------------------
   Example:
   --------------------------------------------------------
      oRef  =  new fontDlg()
      oRef.inspect()
      // assign property values as desired.
      // call methods as desired.
      // call to display the dialog:
      oRef.getFont()
      // make selections.
      // use properties and/or methods.
      oRef.release()
   --------------------------------------------------------
   Notes:
   --------------------------------------------------------
      1) the color and value properties are outputs.
      2) this.iniFile default is:
         a) _app.iniFile when available.
         b) in Runtime _app.exename.ini.
         c) in IDE _dbwinhome + className + .ini.
      3) iniFile assignments without a path default to:
         a) when in runtime = _dbwinhome.
         b) when in ide = set("directory").
      4) selection of OK auto-saves to this.iniFile.
   --------------------------------------------------------
   Revisions:
   --------------------------------------------------------
      February 18, 2011
         1) default ini file adapted to following
            Windows security model and a standard User
            with the use of applicationIni.cc.
      February 10, 2005
         1) renamed all external functions.
      January 05, 2005
         1) fix for the limitSize property not resetting.
         2) changed to default ini when in the IDE (see Notes).
      December 21, 2004
         1) changes to constant names.
         2) standardized the constructor area.
      December 17, 2004
         1) added reset method.
         2) added compatibility for VStruct.prg.
   --------------------------------------------------------
   Properties:
   --------------------------------------------------------
      Name:                Data:    Default:
      -----------------------------------------------------
      bold                 <bool>   false
      color                <char>   "black"
         use as if readOnly.
      colorStyle           <int>    0
         preference for color property format.
         0  =  java color name.
         1  =  dBASE color name.
         2  =  system color name.
         3  =  dBASE hex color.
      effects              <bool>   true
         allow color, underline, strikeout ?
      fontList             <int>    0
         0  =  display screen and printer fonts.
         1  =  display screen fonts.
         2  =  display printer fonts.
         3  =  display scalable fonts.
         4  =  display fonts that are on both the
               screen and printer.
      fontName             <char>   "Arial"
      fontType             <int>    0
         0  =  all types.
         1  =  trueType only.
         2  =  fixedPitch only.
      iniFile              <char>   _app.exename.ini
      italic               <bool>   false
      limitSize            <bool>   false
         restrict font selection to
         >= sizeMin and <= sizeMax.
      messages             <bool>   true
      noFaceSelect         <bool>   false
         display without font selected ?
      noScriptSelect       <bool>   true
         disable the script selection ?
      noSizeSelect         <bool>   false
         display without size selected ?
      noStyleSelect        <bool>   false
         display without style selected ?
      size                 <int>    10
      sizeMax              <int>    72
         requires limitSize.
      sizeMin              <int>    8
         requires limitSize.
      showHelp             <bool>   false
      strikeout            <bool>   false
      underline            <bool>   false
      value                <char>   Arial,10,Swiss
         use as if readOnly.
      winColor             <int>    0
   --------------------------------------------------------
   Methods:
   --------------------------------------------------------
      assignColorNormal(<oRef> GUI object, [<bool> background])
                  assign color selection to object (oRef).
      assignColorHighLight(<oRef> GUI object, [<bool> background])
                  assign color selection to object (oRef).
      assignFont(<oRef> GUI object)
                  assign font selection to object (oRef).
      clearIni()  empty iniFile entries.
      getFont()   return is similar to dBASE getFont() function.
      loadIni()   assign properties from this.iniFile.
      refresh()   updates this.value and this.color properties.
      release()   destroys object, closes procedure files.
      reset()     load all default settings.
      saveIni()   save properties to this.iniFile.
      setColor(<char> dbase color)
                  assign this.winColor with a dBASE color.
      show()      returns true/false (OK/Cancel selection).
*/
// VStruct constants.
#include <VStruct.h>

// choosefont.
#define  CF_SCREENFONTS          0x00000001
#define  CF_PRINTERFONTS         0x00000002
#define  CF_BOTH                 (bitor(CF_SCREENFONTS,CF_PRINTERFONTS))
#define  CF_SHOWHELP             0x00000004
#define  CF_INITTOLOGFONTSTRUCT  0x00000040
#define  CF_EFFECTS              0x00000100
#define  CF_ANSIONLY             0x00000400
#define  CF_SCRIPTSONLY          CF_ANSIONLY
#define  CF_NOVECTORFONTS        0x00000800
#define  CF_NOOEMFONTS           CF_NOVECTORFONTS
#define  CF_NOSIMULATIONS        0x00001000
#define  CF_LIMITSIZE            0x00002000
#define  CF_FIXEDPITCHONLY       0x00004000
#define  CF_WYSIWYG              0x00008000 // must also have CF_SCREENFONTS & CF_PRINTERFONTS
#define  CF_FORCEFONTEXIST       0x00010000
#define  CF_SCALABLEONLY         0x00020000
#define  CF_TTONLY               0x00040000
#define  CF_NOFACESEL            0x00080000
#define  CF_NOSTYLESEL           0x00100000
#define  CF_NOSIZESEL            0x00200000
#define  CF_SELECTSCRIPT         0x00400000
#define  CF_NOSCRIPTSEL          0x00800000
#define  CF_NOVERTFONTS          0x01000000

// font weights.
#define  FW_DONTCARE             0
#define  FW_THIN                 100
#define  FW_EXTRALIGHT           200
#define  FW_LIGHT                300
#define  FW_NORMAL               400
#define  FW_MEDIUM               500
#define  FW_SEMIBOLD             600
#define  FW_BOLD                 700
#define  FW_EXTRABOLD            800
#define  FW_HEAVY                900

// logical font.
#define  OUT_DEFAULT_PRECIS      0
#define  CLIP_DEFAULT_PRECIS     0
#define  DEFAULT_FACENAME        "Arial"
#define  DEFAULT_QUALITY         0
#define  DEFAULT_PITCH           0

// graphics modes.
#define  GM_ADVANCED             2
#define  GM_COMPATIBLE           1

#define  MWT_IDENTITY            1

// global alloc.
#define  GMEM_MOVEABLE           0x00000002
#define  GMEM_ZEROINIT           0x00000040

// device parameters.
#define  HORZSIZE                4  /* Horizontal size in millimeters*/
#define  VERTSIZE                6  /* Vertical size in millimeters*/
#define  HORZRES                 8  /* Horizontal width in pixels*/
#define  VERTRES                 10 /* Vertical height in pixels*/
#define  LOGPIXELSX              88 /* Logical pixels/inch in X*/
#define  LOGPIXELSY              90 /* Logical pixels/inch in Y*/

// GetSysColor identifiers.
// provided by Rich Mueller.
#define  COLOR_SCROLLBAR            0  // The color of a scroll bar's gray area.
#define  COLOR_BACKGROUND           1  // The desktop color.
#define  COLOR_ACTIVECAPTION        2  // The active window title bar color. Windows 98, 2000: The color of the left side of the active window title bar gradient, if the gradient effect is used.
#define  COLOR_INACTIVECAPTION      3  // The color of an inactive window's caption. Windows 98, 2000: The color of the right side of an inactive window's title bar gradient, if the gradient effect is used.
#define  COLOR_MENU                 4  // The background color of menus.
#define  COLOR_WINDOW               5  // The background color of a window.
#define  COLOR_WINDOWFRAME          6  // The color of a window frame.
#define  COLOR_MENUTEXT             7  // The color of menu text.
#define  COLOR_WINDOWTEXT           8  // The color of text in a window.
#define  COLOR_CAPTIONTEXT          9  // The color of window caption text, size boxes, and scroll bar arrow boxes.
#define  COLOR_ACTIVEBORDER         10 // The active window border color.
#define  COLOR_INACTIVEBORDER       11 // The color of an inactive window's border.
#define  COLOR_APPWORKSPACE         12 // The background color of multiple document interface (MDI) windows.
#define  COLOR_HIGHLIGHT            13 // The color used to highlight selected items.
#define  COLOR_HIGHLIGHTTEXT        14 // The color used for the text of highlighted items.
#define  COLOR_BTNFACE              15 // The color of a button's face.
#define  COLOR_BTNSHADOW            16 // The color of a button's shadow area.
#define  COLOR_GRAYTEXT             17 // The color for disabled (grayed-out) text.
#define  COLOR_BTNTEXT              18 // The text color for pushbuttons.
#define  COLOR_INACTIVECAPTIONTEXT  19 // The color of an inactive window's caption text.
#define  COLOR_BTNHIGHLIGHT         20 // The color of a button's highlight area.
#define  COLOR_INFOTEXT             23 // The text color for tooltip controls.
#define  COLOR_INFOBK               24 // The background color for tooltop controls.

// programmer constants.
#define  FONTDLG_DECORATIVE         "Decorative"
#define  FONTDLG_FONTSTYLE          0x00000010
#define  FONTDLG_FONTTYPE           0x00000400
#define  FONTDLG_MODERN             "Modern"
#define  FONTDLG_ROMAN              "Roman"
#define  FONTDLG_SCRIPT             "Script"
#define  FONTDLG_SWISS              "Swiss"
#define  FONTDLG_PUBLISHERINFO      "Rick Miller"
#define  FONTDLG_VERSIONINFO        "11.0218"

// programmer macros.
#define  ghex_color(n)     right("000000" + ItoH(n), 6)
#define  zeros(n)          replicate(chr(0), n)
//---------------------------------------------------------------//
//          constructor.
//---------------------------------------------------------------//
class fontDlg
   this.className =  "fontDlg"

   this.debug     =  false
   this.messages  =  true
   this.versioninfo  =  FONTDLG_VERSIONINFO
   this.publishinfo  =  FONTDLG_PUBLISHERINFO
   this.~runtime  =  iif("run" $ lower(version(0)), true, false)
   this.~procFile =  class::x_GetProcfile()
   this.~procFile2   =  class::x_GetProcfile2()

   if empty(this.~procFile) or empty(this.~procFile2)
      class::release()
   else
      class::x_Protect()
      class::x_InitExterns()
      class::x_Initialize()
   endif

   //------------------------------------------------------------//
   //       end of constructor   -  methods below.
   //------------------------------------------------------------//
   // assign color selection to object.colorHighlight.
   // bBackground is used to determine
   // if the foreGround or backGround color is assigned.
   function assignColorHighLight(oRef, bBackground)
      oRef.colorHighlight  =  ;
               class::x_AssignColor( ;
               "" + oRef.colorHighlight, bBackground)
   return

   //------------------------------------------------------------//
   // assign color selection to object.colorNormal.
   // bBackground is used to determine
   // if the foreGround or backGround color is assigned.
   function assignColorNormal(oRef, bBackground)
      oRef.colorNormal  =  ;
               class::x_AssignColor( ;
               "" + oRef.colorNormal, bBackground)
   return

   //------------------------------------------------------------//
   // assign font selection to object (oRef).
   function assignFont(oRef)
      with(oRef)
         fontName = "" + this.fontName
         fontSize =  int(this.size)
         fontBold    =  iif(this.bold, true, false)
         fontItalic  =  iif(this.italic, true, false)
         fontUnderline  = iif(this.underline, true, false)
         fontStrikeout  = iif(this.strikeout, true, false)
      endwith
   return

   //------------------------------------------------------------//
   // empty the iniFile settings.
   function clearIni
      class::x_WriteIni("FontDlg", "Bold", "")
      class::x_WriteIni("FontDlg", "Color", "")
      class::x_WriteIni("FontDlg", "Fontname", "")
      class::x_WriteIni("FontDlg", "Size", "")
      class::x_WriteIni("FontDlg", "SizeMax", "")
      class::x_WriteIni("FontDlg", "SizeMin", "")
      class::x_WriteIni("FontDlg", "Italic", "")
      class::x_WriteIni("FontDlg", "LimitSize", "")
      class::x_WriteIni("FontDlg", "Strikeout", "")
      class::x_WriteIni("FontDlg", "Underline", "")
   return

   //------------------------------------------------------------//
   // for compatible behavior with the dBASE GetFont() function.
   function getFont
   return iif(class::show(), "" + this.value, "")

   //------------------------------------------------------------//
   // display this in the inspector.
   function inspect
      inspect(this)
   return

   //------------------------------------------------------------//
   // load the last saved values from an ini file.
   function loadIni
      Local cValu, nValu

      cValu =  class::x_ReadIni("FontDlg", "Bold")
      this.bold   =  iif(cValu == "1", true, false)

      cValu =  class::x_ReadIni("FontDlg", "Color")
      this.Color  =  iif(empty(cValu), "black", cValu)

      nValu =  val(class::x_ReadIni("FontDlg", "ColorStyle"))
      this.colorStyle   =  int(mod(nValu, 4))

      cValu =  class::x_ReadIni("FontDlg", "Fontname")
      this.fontName  =  iif(empty(cValu), "Arial", cValu)

      nValu =  val(class::x_ReadIni("FontDlg", "Size"))
      this.Size   =  iif(nValu > 5, nValu, 10)

      nValu =  val(class::x_ReadIni("FontDlg", "SizeMax"))
      this.SizeMax   =  iif(nValu < 73 and nValu > 5, nValu, 72)

      nValu =  val(class::x_ReadIni("FontDlg", "SizeMin"))
      this.SizeMin   =  iif(nValu > 5 and nValu < 73, nValu, 8)

      cValu =  class::x_ReadIni("FontDlg", "Italic")
      this.italic    =  iif(cValu == "1", true, false)

      cValu =  class::x_ReadIni("FontDlg", "LimitSize")
      this.limitSize =  iif(cValu == "1", true, false)

      cValu =  class::x_ReadIni("FontDlg", "Strikeout")
      this.Strikeout =  iif(cValu == "1", true, false)

      cValu =  class::x_ReadIni("FontDlg", "Underline")
      this.underline =  iif(cValu == "1", true, false)
   return

   //------------------------------------------------------------//
   function refresh
      class::x_Refresh()
   return

   //------------------------------------------------------------//
   // destroy object.
   // close this file and it's dependencies.
   function release
      if type("this.~chooseFont") == "O"
         release object this.~chooseFont
         this.~chooseFont  := null
      endif
      if type("this.~logFont") == "O"
         release object this.~logFont
         this.~logFont  := null
      endif
      if type("this.~appIni") == "O"
         this.~appIni.release()
         this.~appIni   := null
      endif
      if type("this.dbaseColors") == "O"
         release object this.dbaseColors
         this.dbaseColors  := null
      endif
      if type("this.dbaseColors16") == "O"
         release object this.dbaseColors16
         this.dbaseColors16   := null
      endif
      if type("this.javaColors") == "O"
         release object this.javaColors
         this.javaColors   := null
      endif
      if type("this.sysColors") == "O"
         release object this.sysColors
         this.sysColors   := null
      endif
      if type("this.sysNames") == "O"
         release object this.sysNames
         this.sysNames  := null
      endif
      class::x_CloseProcedure(this.~procFile)
      class::x_CloseProcedure(this.~procFile2)
      try
         release object this
      catch(exception e)
      endtry
      class::x_CloseProcedure(program(1))
   return

   //------------------------------------------------------------//
   // load all default property values.
   function reset
      this.bold         := false
      this.color        := "black"
      this.colorStyle   :=  0
      this.fontName     := "Arial"
      this.italic       := false
      this.limitSize    := false
      this.messages     := true
      this.size         := 10
      this.sizeMax      := 72
      this.sizeMin      := 8
      this.strikeout    := false
      this.underline    := false
      class::x_LoadDefaultSettings()
      this.iniFile      := null
      this.iniFile      := this.~appIni.getFilename()
      this.winColor     := class::x_GetColor(this.Color)
      class::refresh()
   return

   //------------------------------------------------------------//
   // save the dialog properties to  this.iniFile.
   function saveIni
      if this.~appIni.fileCheck(this.iniFile)
         class::x_WriteIni("FontDlg", "Bold", iif(this.bold, "1", "0"))
         class::x_WriteIni("FontDlg", "Color", "" + this.Color)
         class::x_WriteIni("FontDlg", "ColorStyle", "" + this.colorStyle)
         class::x_WriteIni("FontDlg", "Fontname", "" + this.fontName)
         class::x_WriteIni("FontDlg", "Size", "" + this.Size)
         class::x_WriteIni("FontDlg", "SizeMax", "" + this.SizeMax)
         class::x_WriteIni("FontDlg", "SizeMin", "" + this.SizeMin)
         class::x_WriteIni("FontDlg", "Italic", iif(this.italic, "1", "0"))
         class::x_WriteIni("FontDlg", "LimitSize", iif(this.limitSize, "1", "0"))
         class::x_WriteIni("FontDlg", "Strikeout", iif(this.strikeout, "1", "0"))
         class::x_WriteIni("FontDlg", "Underline", iif(this.underline, "1", "0"))
      endif
   return

   //------------------------------------------------------------//
   // assign this.winColor from a dBASE color.
   function setColor(cColr)
      this.winColor  := class::x_GetColor("" + cColr)
   return

   //------------------------------------------------------------//
   // display the dialog.
   function show
      // update the property objects for any changes.
      class::x_RefreshChooseFont()
      class::x_RefreshLogFont()

      Local bRet, hMem, nDc, nHwnd, nSize, oFont, oLog

      // create new structures for the dialog.
      oFont =  class::x_CreateChooseFont()
      oLog  =  class::x_CreateLogFont()

      // place this object's properties into the structures.
      oLog.string    =  this.~logFont.string
      oFont.string   =  this.~chooseFont.string

      // get the active window handle.
      nHwnd =  RMM_GetActiveWindow()
      // get the device context.
      nDc   =  RMM_GetDC(nHwnd)

      // copy the logFont structure into global memory.
      nSize =  oLog.getSize()
      // allocate memory.
      hMem  =  RMM_GlobalAlloc(GMEM_MOVEABLE + GMEM_ZEROINIT, nSize)
      // get pointer and increment lock count.
      pMem  =  RMM_GlobalLock(hMem)
      // copy to memory at pMem from oLog.
      RMM_CopyMemoryTo(pMem, oLog.string, nSize)

      // destroy oLog.
      release object oLog
      oLog  := null

      // set the chooseFont members.
      oFont.setMem('hwndOwner', nHwnd)
      oFont.setMem('hDC', nDc)
      oFont.setMem('lpLogFont', pMem)

      // open the dialog.
      bRet  =  RMM_ChooseFont(oFont)

      if bRet
         // user selected the OK button.
         // update this with the dialog selections made.

         // get the memory size (may have changed).
         nSize := RMM_GlobalSize(hMem)
         // create a new string to copy the memory to.
         oLog  := zeros(nSize)
         // copy from memory at pMem to oLog.
         RMM_CopyMemoryFrom(oLog, pMem, nSize)

         this.~logFont.string    := oLog.string
         this.~chooseFont.string := oFont.string
         class::x_Update()
      endif

      // decrement the memory lock count.
      RMM_GlobalUnlock(hMem)
      // free the global memory.
      RMM_GlobalFree(hMem)
      // release the device context.
      RMM_ReleaseDC(nHwnd, nDc)

      // destroy oFont.
      release object oFont
      oFont := null
   return   iif(bRet, true, false)

   //------------------------------------------------------------//
   // return a dBASE colorNormal value.
   // bBack determines whether to place this.color
   // in the foreGround or backGround position.
   function x_AssignColor(cColr, bBack)
      Local cBg, cFg, nPos
      cFg   =  ""
      cBg   =  ""
      nPos  =  cColr.indexof("/")
      if bBack
         cBg   := "" + this.color
         if not nPos == -1
            cFg   := cColr.substring(0, nPos + 1)
         endif
      else
         cFg   := "" + this.color
         if not nPos == -1
            cBg   := cColr.substring(nPos, cColr.length)
         endif
      endif
   return   cFg + cBg

   //------------------------------------------------------------//
   // try and close a procedure file.
   function x_CloseProcedure(cProc)
      try
         close procedure (cProc)
      catch(exception e)
      endtry
   return

   //------------------------------------------------------------//
   // create a chooseFont structure.
   function x_CreateChooseFont
      Local oFont
      oFont =  FONTDLG_Structures::chooseFontStructure()
      // assign default values to members.
      oFont.setMem('rgbColors', (this.winColor))
      oFont.setMem('nFontType', FONTDLG_FONTTYPE + FONTDLG_FONTSTYLE)
   return   oFont

   //------------------------------------------------------------//
   // create a logFont structure.
   function x_CreateLogFont
      Local oLog
      oLog  =  FONTDLG_Structures::logFontStructure()
      // assign default values to members.
      oLog.setMem('lfHeight',          -13)
      oLog.setMem('lfWeight',          FW_NORMAL)
      oLog.setMem('lfOutPrecision',    OUT_DEFAULT_PRECIS)
      oLog.setMem('lfClipPrecision',   CLIP_DEFAULT_PRECIS)
      oLog.setMem('lfQuality',         DEFAULT_QUALITY)
      oLog.setMem('lfPitchAndFamily',  DEFAULT_PITCH)
      oLog.setMem('lfFaceName',        DEFAULT_FACENAME + chr(0))
   return   oLog

   //------------------------------------------------------------//
   function x_ErrorMessage(cBody)
      if lower("" + this.messages) == "true"
         msgbox(cBody, "  fontDlg " + this.version, 64)
      endif
   return

   //------------------------------------------------------------//
   // return a winColor <int>.
   function x_GetColor(cValu)
      Local cVal, i, nRet
      cVal  =  cValu.toLowerCase()
      if not cVal.indexof("/") == -1
         cVal  := cVal.left(cVal.indexof("/"))
      endif
      nRet  =  0
      if this.sysNames.iskey(cVal)
         nRet  := RMM_GetSysColor(this.sysColors[cVal])
      elseif this.javaColors.iskey(cVal)
         nRet  := this.javaColors[cVal]
      elseif this.dbaseColors.iskey(cVal)
         nRet  := this.dbaseColors[cVal]
      elseif cVal.left(2) == "0x"
         nRet  := HtoI(cVal.substring(2, cVal.length))
      endif
   return   int(nRet)

   //------------------------------------------------------------//
   // return the VStruct file that will be used.
   function x_GetProcfile
      Local cRet
      cRet  =  ""
      try
         set procedure to VStruct.cc additive
         cRet  := "VStruct.cc"
      catch(exception e)
         try
            set procedure to VStruct32Ex.cc additive
            cRet  := "VStruct32Ex.cc"
         catch(exception e)
            try
               set procedure to VStruct.prg additive
               cRet  := "VStruct.prg"
            catch(exception e)
               try
                  set procedure to :dUFLP:VStruct.prg additive
                  cRet  := ":dUFLP:VStruct.prg"
               catch(exception e)
                  class::x_ErrorMessage("VStruct files are missing.")
               endtry
            endtry
         endtry
      endtry
   return   cRet

   //------------------------------------------------------------//
   // return the applicationIni.cc file that will be used.
   function x_GetProcfile2
      Local cRet
      cRet  =  ""
      try
         set procedure to applicationIni.cc additive
         cRet  := "applicationIni.cc"
      catch(exception e)
         try
            set procedure to :dUFLP:applicationIni.cc additive
            cRet  := ":dUFLP:applicationIni.cc"
         catch(exception e)
            class::x_ErrorMessage( ;
                     "applicationIni file is missing.")
         endtry
      endtry
   return   cRet

   //------------------------------------------------------------//
   // initialize external functions.
   function x_InitExterns
      //---------------------------------------------------------//
      //    font functions.
      //---------------------------------------------------------//
      if not type("RMM_DPtoLP") == "FP"
         extern   CLOGICAL ;
                  RMM_DPtoLP(;
                  CHANDLE,;
                  CPTR,;
                  CINT) ;
                  gdi32 from "DPtoLP"
      endif
      if not type("RMM_GetDeviceCaps") == "FP"
         extern   CINT ;
                  RMM_GetDeviceCaps(;
                  CHANDLE,;
                  CINT) ;
                  gdi32 from "GetDeviceCaps"
      endif
      if not type("RMM_GetTextCharset") == "FP"
         extern   CINT ;
                  RMM_GetTextCharset(;
                  CHANDLE) ;
                  gdi32 from "GetTextCharset"
      endif
      if not type("RMM_ModifyWorldTransform") == "FP"
         extern   CLOGICAL ;
                  RMM_ModifyWorldTransform(;
                  CHANDLE,;
                  CPTR,;
                  CULONG) ;
                  gdi32 from "ModifyWorldTransform"
      endif
      if not type("RMM_RestoreDC") == "FP"
         extern   CLOGICAL ;
                  RMM_RestoreDC(;
                  CHANDLE,;
                  CINT) ;
                  gdi32 from "RestoreDC"
      endif
      if not type("RMM_SaveDC") == "FP"
         extern   CINT ;
                  RMM_SaveDC(;
                  CHANDLE) ;
                  gdi32 from "SaveDC"
      endif
      if not type("RMM_SetGraphicsMode") == "FP"
         extern   CINT ;
                  RMM_SetGraphicsMode(;
                  CHANDLE,;
                  CINT) ;
                  gdi32 from "SetGraphicsMode"
      endif
      if not type("RMM_SetViewportOrgEx") == "FP"
         extern   CLOGICAL ;
                  RMM_SetViewportOrgEx(;
                  CHANDLE,;
                  CINT,;
                  CINT,;
                  CPTR) ;
                  gdi32 from "SetViewportOrgEx"
      endif
      if not type("RMM_SetWindowOrgEx") == "FP"
         extern   CLOGICAL ;
                  RMM_SetWindowOrgEx(;
                  CHANDLE, ;
                  CINT,;
                  CINT,;
                  CPTR) ;
                  gdi32 from "SetWindowOrgEx"
      endif
      //---------------------------------------------------------//
      //    dialog functions.
      //---------------------------------------------------------//
      if not type("RMM_ChooseFont") == "FP"
         extern   CLOGICAL ;
                  RMM_ChooseFont(;
                  CPTR) ;
                  comdlg32 from "ChooseFontA"
      endif
      if not type("RMM_CopyMemoryFrom") == "FP"
         extern   CVOID ;
                  RMM_CopyMemoryFrom(;
                  CPTR,;
                  CPTR,;
                  CULONG) ;
                  kernel32 from "RtlMoveMemory"
      endif
      if not type("RMM_CopyMemoryTo") == "FP"
         extern   CVOID ;
                  RMM_CopyMemoryTo(;
                  CLONG,;
                  CPTR,;
                  CULONG) ;
                  kernel32 from "RtlMoveMemory"
      endif
      if not type("RMM_GetActiveWindow") == "FP"
         extern   CHANDLE ;
                  RMM_GetActiveWindow(;
                  CVOID) ;
                  user32 from "GetActiveWindow"
      endif
      if not type("RMM_GetDC") == "FP"
         extern   CHANDLE ;
                  RMM_GetDC(;
                  CHANDLE) ;
                  user32 from "GetDC"
      endif
      if not type("RMM_GlobalAlloc") == "FP"
         extern   CHANDLE ;
                  RMM_GlobalAlloc(;
                  CUINT,;
                  CULONG) ;
                  kernel32 from "GlobalAlloc"
      endif
      if not type("RMM_GlobalFree") == "FP"
         extern   CHANDLE ;
                  RMM_GlobalFree(;
                  CHANDLE) ;
                  kernel32 from "GlobalFree"
      endif
      if not type("RMM_GlobalLock") == "FP"
         extern   CLONG ;
                  RMM_GlobalLock(;
                  CHANDLE) ;
                  kernel32 from "GlobalLock"
      endif
      if not type("RMM_GlobalSize") == "FP"
         extern   CULONG ;
                  RMM_GlobalSize(;
                  CHANDLE) ;
                  kernel32 from 'GlobalSize'
      endif
      if not type("RMM_GlobalUnLock") == "FP"
         extern   CLOGICAL ;
                  RMM_GlobalUnLock(;
                  CHANDLE) ;
                  kernel32 from "GlobalUnlock"
      endif
      if not type("RMM_ReleaseDC") == "FP"
         extern   CINT ;
                  RMM_ReleaseDC(;
                  CHANDLE,;
                  CHANDLE) ;
                  user32 from "ReleaseDC"
      endif
      //---------------------------------------------------------//
      //    misc. functions.
      //---------------------------------------------------------//
      if not type("RMM_GetSysColor") == "FP"
         extern   CULONG ;
                  RMM_GetSysColor(;
                  CINT) ;
                  user32 from "GetSysColor"
      endif
   return

   //------------------------------------------------------------//
   function x_Initialize
      class::x_LoadColors()
      this.~appIni   =  new applicationIni()
      this.iniFile   =  this.~appIni.value
      class::loadIni()
      class::x_LoadDefaultSettings()
      this.winColor     =  class::x_GetColor(this.Color)
      this.~chooseFont  =  class::x_CreateChooseFont()
      this.~logFont     =  class::x_CreateLogFont()
      class::refresh()
   return

   //------------------------------------------------------------//
   function x_LoadColors
      this.dBASEColors  =  new assocArray()
      this.dbaseColors["b"]   =   8388608    // navy
      this.dbaseColors["b+"]  =  16711680    // blue
      this.dbaseColors["b*"]  =  16711680    // blue
      this.dbaseColors["+b"]  =  16711680    // blue
      this.dbaseColors["*b"]  =  16711680    // blue
      this.dbaseColors["g"]   =     32768    // dark green
      this.dbaseColors["g+"]  =     65280    // lime
      this.dbaseColors["g*"]  =     65280    // lime
      this.dbaseColors["+g"]  =     65280    // lime
      this.dbaseColors["*g"]  =     65280    // lime
      this.dbaseColors["n"]   =         0    // black
      this.dbaseColors["n+"]  =   8421504    // gray
      this.dbaseColors["n*"]  =   8421504    // gray
      this.dbaseColors["+n"]  =   8421504    // gray
      this.dbaseColors["*n"]  =   8421504    // gray
      this.dbaseColors["r"]   =       128    // maroon
      this.dbaseColors["r+"]  =       255    // red
      this.dbaseColors["r*"]  =       255    // red
      this.dbaseColors["+r"]  =       255    // red
      this.dbaseColors["*r"]  =       255    // red
      this.dbaseColors["w"]   =  12632256    // silver
      this.dbaseColors["w+"]  =  16777215    // white
      this.dbaseColors["w*"]  =  16777215    // white
      this.dbaseColors["+w"]  =  16777215    // white
      this.dbaseColors["*w"]  =  16777215    // white
      this.dbaseColors["bg"]  =   8421376    // teal
      this.dbaseColors["bg+"] =  16776960    // cyan
      this.dbaseColors["bg*"] =  16776960    // cyan
      this.dbaseColors["+bg"] =  16776960    // cyan
      this.dbaseColors["b+g"] =  16776960    // cyan
      this.dbaseColors["*bg"] =  16776960    // cyan
      this.dbaseColors["b*g"] =  16776960    // cyan
      this.dbaseColors["br"]  =   8388736    // purple
      this.dbaseColors["br+"] =  16711935    // magenta
      this.dbaseColors["br*"] =  16711935    // magenta
      this.dbaseColors["+br"] =  16711935    // magenta
      this.dbaseColors["*br"] =  16711935    // magenta
      this.dbaseColors["b+r"] =  16711935    // magenta
      this.dbaseColors["b*r"] =  16711935    // magenta
      this.dbaseColors["gb"]  =   8421376    // teal
      this.dbaseColors["gb+"] =  16776960    // cyan
      this.dbaseColors["gb*"] =  16776960    // cyan
      this.dbaseColors["+gb"] =  16776960    // cyan
      this.dbaseColors["*gb"] =  16776960    // cyan
      this.dbaseColors["g+b"] =  16776960    // cyan
      this.dbaseColors["g*b"] =  16776960    // cyan
      this.dbaseColors["gr"]  =     32896    // olive
      this.dbaseColors["gr+"] =     65535    // yellow
      this.dbaseColors["gr*"] =     65535    // yellow
      this.dbaseColors["+gr"] =     65535    // yellow
      this.dbaseColors["*gr"] =     65535    // yellow
      this.dbaseColors["g+r"] =     65535    // yellow
      this.dbaseColors["g*r"] =     65535    // yellow
      this.dbaseColors["rb"]  =   8388736    // purple
      this.dbaseColors["rb+"] =  16711935    // magenta
      this.dbaseColors["rb*"] =  16711935    // magenta
      this.dbaseColors["+rb"] =  16711935    // magenta
      this.dbaseColors["*rb"] =  16711935    // magenta
      this.dbaseColors["r+b"] =  16711935    // magenta
      this.dbaseColors["r*b"] =  16711935    // magenta
      this.dbaseColors["rg"]  =     32896    // olive
      this.dbaseColors["rg+"] =     65535    // yellow
      this.dbaseColors["rg*"] =     65535    // yellow
      this.dbaseColors["+rg"] =     65535    // yellow
      this.dbaseColors["*rg"] =     65535    // yellow
      this.dbaseColors["r+g"] =     65535    // yellow
      this.dbaseColors["r*g"] =     65535    // yellow

      this.dBASEColors16  =  new assocArray()
      this.dbaseColors16["b"]    =  this.dbaseColors["b"]
      this.dbaseColors16["b+"]   =  this.dbaseColors["b+"]
      this.dbaseColors16["g"]    =  this.dbaseColors["g"]
      this.dbaseColors16["g+"]   =  this.dbaseColors["g+"]
      this.dbaseColors16["n"]    =  this.dbaseColors["n"]
      this.dbaseColors16["n+"]   =  this.dbaseColors["n+"]
      this.dbaseColors16["r"]    =  this.dbaseColors["r"]
      this.dbaseColors16["r+"]   =  this.dbaseColors["r+"]
      this.dbaseColors16["w"]    =  this.dbaseColors["w"]
      this.dbaseColors16["w+"]   =  this.dbaseColors["w+"]
      this.dbaseColors16["gb"]   =  this.dbaseColors["gb"]
      this.dbaseColors16["gb+"]  =  this.dbaseColors["gb+"]
      this.dbaseColors16["gr"]   =  this.dbaseColors["gr"]
      this.dbaseColors16["gr+"]  =  this.dbaseColors["gr+"]
      this.dbaseColors16["rb"]   =  this.dbaseColors["rb"]
      this.dbaseColors16["rb+"]  =  this.dbaseColors["rb+"]

      this.javaColors   =  new assocArray()
      this.javaColors["aliceblue"]           =  16775408
      this.javaColors["antiquewhite"]        =  14150650
      this.javaColors["aqua"]                =  16776960
      this.javaColors["aquamarine"]          =  13959039
      this.javaColors["azure"]               =  16777200
      this.javaColors["beige"]               =  14480885
      this.javaColors["bisque"]              =  12903679
      this.javaColors["black"]               =         0
      this.javaColors["blanchedalmond"]      =  13495295
      this.javaColors["blue"]                =  16711680
      this.javaColors["blueviolet"]          =  14822282
      this.javaColors["brown"]               =   2763429
      this.javaColors["burlywood"]           =   8894686
      this.javaColors["cadetblue"]           =  10526303
      this.javaColors["chartreuse"]          =     65407
      this.javaColors["chocolate"]           =   1993170
      this.javaColors["coral"]               =   5275647
      this.javaColors["cornflowerblue"]      =  15570276
      this.javaColors["cornsilk"]            =  14481663
      this.javaColors["crimson"]             =   3937500
      this.javaColors["cyan"]                =  16776960
      this.javaColors["darkblue"]            =   9109504
      this.javaColors["darkcyan"]            =   9145088
      this.javaColors["darkgoldenrod"]       =    755384
      this.javaColors["darkgray"]            =  11119017
      this.javaColors["darkgreen"]           =     25600
      this.javaColors["darkkhaki"]           =   7059389
      this.javaColors["darkmagenta"]         =   9109643
      this.javaColors["darkolivegreen"]      =   3107669
      this.javaColors["darkorange"]          =     36095
      this.javaColors["darkorchid"]          =  13382297
      this.javaColors["darkred"]             =       139
      this.javaColors["darksalmon"]          =   8034025
      this.javaColors["darkseagreen"]        =   9419919
      this.javaColors["darkslateblue"]       =   9125192
      this.javaColors["darkslategray"]       =   5197615
      this.javaColors["darkturquoise"]       =  13749760
      this.javaColors["darkviolet"]          =  13828244
      this.javaColors["deeppink"]            =   9639167
      this.javaColors["deepskyblue"]         =  16760576
      this.javaColors["dimgray"]             =   6908265
      this.javaColors["dodgerblue"]          =  16748574
      this.javaColors["firebrick"]           =   2237106
      this.javaColors["floralwhite"]         =  15792895
      this.javaColors["forestgreen"]         =   2263842
      this.javaColors["fuchsia"]             =  16711935
      this.javaColors["gainsboro"]           =  14474460
      this.javaColors["ghostwhite"]          =  16775416
      this.javaColors["gold"]                =     55295
      this.javaColors["goldenrod"]           =   2139610
      this.javaColors["gray"]                =   8421504
      this.javaColors["green"]               =     32768
      this.javaColors["greenyellow"]         =   3145645
      this.javaColors["honeydew"]            =  15794160
      this.javaColors["hotpink"]             =  11823615
      this.javaColors["indianred"]           =   6053069
      this.javaColors["indigo"]              =   8519755
      this.javaColors["ivory"]               =  15794175
      this.javaColors["khaki"]               =   9234160
      this.javaColors["lavender"]            =  16443110
      this.javaColors["lavenderblush"]       =  16118015
      this.javaColors["lawngreen"]           =     64636
      this.javaColors["lemonchiffon"]        =  13499135
      this.javaColors["lightblue"]           =  15128749
      this.javaColors["lightcoral"]          =   8421616
      this.javaColors["lightcyan"]           =  16777184
      this.javaColors["lightgoldenrodyellow"]=  13826810
      this.javaColors["lightgreen"]          =   9498256
      this.javaColors["lightgrey"]           =  13882323
      this.javaColors["lightpink"]           =  12695295
      this.javaColors["lightsalmon"]         =   8036607
      this.javaColors["lightseagreen"]       =  11186720
      this.javaColors["lightskyblue"]        =  16436871
      this.javaColors["lightslategray"]      =  10061943
      this.javaColors["lightsteelblue"]      =  14599344
      this.javaColors["lightyellow"]         =  14745599
      this.javaColors["lime"]                =     65280
      this.javaColors["limegreen"]           =   3329330
      this.javaColors["linen"]               =  15134970
      this.javaColors["magenta"]             =  16711935
      this.javaColors["maroon"]              =       128
      this.javaColors["mediumaquamarine"]    =  11193702
      this.javaColors["mediumblue"]          =  13434880
      this.javaColors["mediumorchid"]        =  13850042
      this.javaColors["mediumpurple"]        =  14381203
      this.javaColors["mediumseagreen"]      =   7451452
      this.javaColors["mediumslateblue"]     =  15624315
      this.javaColors["mediumspringgreen"]   =  10156544
      this.javaColors["mediumturquoise"]     =  13422920
      this.javaColors["mediumvioletred"]     =   8721863
      this.javaColors["midnightblue"]        =   7346457
      this.javaColors["mintcream"]           =  16449525
      this.javaColors["mistyrose"]           =  14804223
      this.javaColors["moccasin"]            =  11920639
      this.javaColors["navajowhite"]         =  11394815
      this.javaColors["navy"]                =   8388608
      this.javaColors["oldlace"]             =  15136253
      this.javaColors["olive"]               =     32896
      this.javaColors["olivedrab"]           =   2330219
      this.javaColors["orange"]              =     42495
      this.javaColors["orangered"]           =     17919
      this.javaColors["orchid"]              =  14053594
      this.javaColors["palegoldenrod"]       =  11200750
      this.javaColors["palegreen"]           =  10025880
      this.javaColors["paleturquoise"]       =  15658671
      this.javaColors["palevioletred"]       =   9662683
      this.javaColors["papayawhip"]          =  14020607
      this.javaColors["peachpuff"]           =  12180223
      this.javaColors["peru"]                =   4163021
      this.javaColors["pink"]                =  13353215
      this.javaColors["plum"]                =  14524637
      this.javaColors["powderblue"]          =  15130800
      this.javaColors["purple"]              =   8388736
      this.javaColors["red"]                 =       255
      this.javaColors["rosybrown"]           =   9408444
      this.javaColors["royalblue"]           =  14772545
      this.javaColors["saddlebrown"]         =   1262987
      this.javaColors["salmon"]              =   7504122
      this.javaColors["sandybrown"]          =   6333684
      this.javaColors["seagreen"]            =   5737262
      this.javaColors["seashell"]            =  15660543
      this.javaColors["sienna"]              =   2970272
      this.javaColors["silver"]              =  12632256
      this.javaColors["skyblue"]             =  15453831
      this.javaColors["slateblue"]           =  13458026
      this.javaColors["slategray"]           =   9470064
      this.javaColors["snow"]                =  16448255
      this.javaColors["springgreen"]         =   8388352
      this.javaColors["steelblue"]           =  11829830
      this.javaColors["tan"]                 =   9221330
      this.javaColors["teal"]                =   8421376
      this.javaColors["thistle"]             =  14204888
      this.javaColors["tomato"]              =   4678655
      this.javaColors["turquoise"]           =  13688896
      this.javaColors["violet"]              =  15631086
      this.javaColors["wheat"]               =  11788021
      this.javaColors["white"]               =  16777215
      this.javaColors["whitesmoke"]          =  16119285
      this.javaColors["yellow"]              =     65535
      this.javaColors["yellowgreen"]         =   3329434

      this.sysColors =  new assocArray()
      this.sysColors["activeborder"]   =  COLOR_ACTIVEBORDER
      this.sysColors["activecaption"]  =  COLOR_ACTIVECAPTION
      this.sysColors["appworkspace"]   =  COLOR_APPWORKSPACE
      this.sysColors["background"]     =  COLOR_BACKGROUND
      this.sysColors["btnface"]        =  COLOR_BTNFACE
      this.sysColors["btnhighlight"]   =  COLOR_BTNHIGHLIGHT
      this.sysColors["btnshadow"]      =  COLOR_BTNSHADOW
      this.sysColors["btntext"]        =  COLOR_BTNTEXT
      this.sysColors["captiontext"]    =  COLOR_CAPTIONTEXT
      this.sysColors["graytext"]       =  COLOR_GRAYTEXT
      this.sysColors["highlight"]      =  COLOR_HIGHLIGHT
      this.sysColors["highlighttext"]  =  COLOR_HIGHLIGHTTEXT
      this.sysColors["inactiveborder"] =  COLOR_INACTIVEBORDER
      this.sysColors["inactivecaption"]      =  COLOR_INACTIVECAPTION
      this.sysColors["inactivecaptiontext"]  =  COLOR_INACTIVECAPTIONTEXT
      this.sysColors["infobk"]         =  COLOR_INFOBK
      this.sysColors["infotext"]       =  COLOR_INFOTEXT
      this.sysColors["menu"]           =  COLOR_MENU
      this.sysColors["menutext"]       =  COLOR_MENUTEXT
      this.sysColors["scrollbar"]      =  COLOR_SCROLLBAR
      this.sysColors["window"]         =  COLOR_WINDOW
      this.sysColors["windowframe"]    =  COLOR_WINDOWFRAME
      this.sysColors["windowtext"]     =  COLOR_WINDOWTEXT

      this.sysNames  =  new assocArray()
      this.sysNames["activeborder"]    =  "ActiveBorder"
      this.sysNames["activecaption"]   =  "ActiveCaption"
      this.sysNames["appworkspace"]    =  "AppWorkspace"
      this.sysNames["background"]      =  "Background"
      this.sysNames["btnface"]         =  "BtnFace"
      this.sysNames["btnhighlight"]    =  "BtnHighlight"
      this.sysNames["btnshadow"]       =  "BtnShadow"
      this.sysNames["btntext"]         =  "BtnText"
      this.sysNames["captiontext"]     =  "CaptionText"
      this.sysNames["graytext"]        =  "GrayText"
      this.sysNames["highlight"]       =  "Highlight"
      this.sysNames["highlighttext"]   =  "HighlightText"
      this.sysNames["inactiveborder"]  =  "InactiveBorder"
      this.sysNames["inactivecaption"]       =  "InactiveCaption"
      this.sysNames["inactivecaptiontext"]   =  "InactiveCaptionText"
      this.sysNames["infobk"]          =  "InfoBk"
      this.sysNames["infotext"]        =  "InfoText"
      this.sysNames["menu"]            =  "Menu"
      this.sysNames["menutext"]        =  "MenuText"
      this.sysNames["scrollbar"]       =  "Scrollbar"
      this.sysNames["window"]          =  "Window"
      this.sysNames["windowframe"]     =  "WindowFrame"
      this.sysNames["windowtext"]      =  "WindowText"
   return

   //------------------------------------------------------------//
   // load default dialog settings.
   function x_LoadDefaultSettings
      this.effects         =  true
      this.fontList        =  0
      this.fontType        =  0
      this.noFaceSelect    =  false
      this.noScriptSelect  =  true
      this.noSizeSelect    =  false
      this.noStyleSelect   =  false
      this.showHelp        =  false
   return

   //------------------------------------------------------------//
   function x_Protect

      // protected properties.
      protect  ~appIni, ~chooseFont, ~logFont,;
               ~procFile, ~procFile2, ~runtime,;
               dbaseColors, dbaseColors16, javaColors, sysNames,;
               sysColors

      // protected methods.
      protect  x_AssignColor, x_CloseProcedure,;
               x_CreateChooseFont, x_CreateLogFont,;
               x_ErrorMessage, x_GetColor, x_GetProcfile,;
               x_GetProcfile2, x_InitExterns, x_Initialize,;
               x_LoadColors, x_LoadDefaultSettings, x_Protect,;
               x_ReadIni, x_Refresh, x_RefreshChooseFont,;
               x_RefreshColor, x_RefreshLogFont, x_Update,;
               x_WriteIni
   return

   //------------------------------------------------------------//
   // read a value from an ini file.
   function x_ReadIni(cSect, cItem)
   return   this.~appIni.getValue(;
            cSect,;
            cItem,;
            this.iniFile)

   //------------------------------------------------------------//
   // assign this.value and this.color.
   function x_Refresh
      Local cBold, cStyl, nStyl
      cBold =  iif(this.bold, "B", "")
      cBold += iif(this.italic, "I", "")
      cBold += iif(this.underline, "U", "")
      cBold += iif(this.strikeout, "S", "")
      if not empty(cBold)
         cBold := "," + cBold
      endif
      // roman = 18, swiss = 34, modern = 49,
      // script = 66, decorative = 82.
      nStyl =  this.~logFont.getMem('lfPitchAndFamily')
      if nStyl < 18
         cStyl =  ""
      elseif nStyl < 34
         cStyl =  FONTDLG_ROMAN
      elseif nStyl < 49
         cStyl =  FONTDLG_SWISS
      elseif nStyl < 66
         cStyl =  FONTDLG_MODERN
      elseif nStyl < 82
         cStyl =  FONTDLG_SCRIPT
      else
         cStyl =  FONTDLG_DECORATIVE
      endif
      if not empty(cStyl)
         cStyl := "," + cStyl
      endif
      this.value  =  ;
               this.fontName + "," +;
               this.Size +;
               cBold + cStyl
      this.color  =  class::x_RefreshColor()
   return

   //------------------------------------------------------------//
   // refresh the this.~chooseFont
   function x_RefreshChooseFont
      Local oFont, nFlag, nList, nType
      oFont =  class::x_CreateChooseFont()
      nFlag =  CF_INITTOLOGFONTSTRUCT

      nList =  int(mod(val("" + this.fontList), 5))
      if nList == 4
         nFlag += CF_BOTH + CF_SCALABLEONLY + CF_WYSIWYG
      elseif nList == 3
         nFlag += CF_BOTH + CF_SCALABLEONLY
      elseif nList == 2
         nFlag += CF_SCREENFONTS
      elseif nList == 1
         nFlag += CF_PRINTERFONTS
      else
         nFlag += CF_BOTH
      endif

      nType =  int(mod(val("" + this.fontType), 3))
      if nType == 1
         nFlag += CF_TTONLY
      elseif nType == 2
         nFlag += CF_FIXEDPITCHONLY
      endif

      if this.noFaceSelect
         nFlag += CF_NOFACESEL
      endif
      if this.noScriptSelect
         nFlag += CF_NOSCRIPTSEL
      endif
      if this.noSizeSelect
         nFlag += CF_NOSIZESEL
      endif
      if this.noStyleSelect
         nFlag += CF_NOSTYLESEL
      endif

      if this.showHelp
         nFlag += CF_SHOWHELP
      endif

      if this.effects
         nFlag += CF_EFFECTS
      endif

      if this.limitSize
         nFlag += CF_LIMITSIZE
         oFont.setMem('nSizeMin',  (this.SizeMin))
         oFont.setMem('nSizeMax',  (this.SizeMax))
      endif
      oFont.setMem('Flags',     nFlag)
      oFont.setMem('rgbColors', (this.winColor))

      this.~chooseFont.string := oFont.string

      release object oFont
      oFont := null

   return

   //------------------------------------------------------------//
   // refresh this.color.
   function x_RefreshColor
      Local cKey, cRet, cOrd, i, ii, nChck, nColr
      nChck =  int(mod(val("" + this.colorStyle), 4))
      nColr =  int(this.winColor)
      if nChck == 1
         cOrd  =  "1,0,2,3"
      elseif nChck == 2
         cOrd  =  "2,0,1,3"
      elseif nChck == 3
         cOrd  =  "3,0,1,2"
      else
         cOrd  =  "0,1,2,3"
      endif
      cRet  =  "BtnText"
      for i = 0 to 3
         nChck := int(val(cOrd.substring(i * 2, i * 2 + 1)))
         if nChck == 3
            // dBase hex color.
            cRet  := "0x" + ghex_color(nColr)
            i  := 6
         elseif nChck == 2
            // sys color name.
            cKey  =  this.sysColors.firstKey
            for ii = 1 to this.sysColors.count()
               if RMM_GetSysColor(this.sysColors[cKey]) == nColr
                  cRet  := this.sysNames[cKey]
                  ii := 400
                  i  := 6
               else
                  cKey  := this.sysColors.nextKey(cKey)
               endif
            endfor
         elseif nChck == 1
            // dBASE color name.
            cKey  =  this.dbaseColors16.firstKey
            for ii = 1 to this.dbaseColors16.count()
               if this.dbaseColors16[cKey] == nColr
                  cRet  := cKey
                  ii := 400
                  i  := 6
               else
                  cKey  := this.dbaseColors16.nextKey(cKey)
               endif
            endfor
         elseif nChck == 0
            // java color name.
            cKey  =  this.javaColors.firstKey
            for ii = 1 to this.javaColors.count()
               if this.javaColors[cKey] == nColr
                  cRet  := cKey
                  ii := 400
                  i  := 6
               else
                  cKey  := this.javaColors.nextKey(cKey)
               endif
            endfor
         endif
      endfor
   return   cRet

   //------------------------------------------------------------//
   // refresh this.~logFont.
   function x_RefreshLogFont
      Local nChr, nDc, nHwnd, nSize, nXdpi, nYdpi, oLog, oPt
      oLog  =  class::x_CreateLogFont()
      oLog.string =  (this.~logFont.string)
      nHwnd =  RMM_GetActiveWindow()
      nSize =  (10 * this.Size)
      nDc   =  RMM_GetDC(nHwnd)
      nChr  =  RMM_GetTextCharset(nDc)
      RMM_SaveDC(nDc)
      // update the dc.
      RMM_SetGraphicsMode(nDc, GM_ADVANCED)
      RMM_ModifyWorldTransform(nDc, 0, MWT_IDENTITY)
      RMM_SetViewportOrgEx(nDc, 0, 0, 0)
      RMM_SetWindowOrgEx(nDc, 0, 0, 0)

      // get the Decimal values for the dc.
      nXdpi =  RMM_GetDeviceCaps(nDc, LOGPIXELSX)
      nYdpi =  RMM_GetDeviceCaps(nDc, LOGPIXELSY)

      // create a point structure that will be Decimal.
      oPt   =  FONTDLG_Structures::pointStructure()
      // fill the point structure.
      oPt.setMem('x', int(ceiling(nSize * nXdpi / 72 + 0.5)))
      oPt.setMem('y', int(ceiling(nSize * nYdpi / 72 + 0.5)))

      // convert the point structure from Decimal to Logical values
      // so it can be placed into the logFont structure.
      RMM_DPtoLP(nDc, oPt, 1)

      // update the lfHeight, lfWeight, lfItalic,
      // lfUnderline, lfStrikeout, lfFaceName, lfCharset members.
      oLog.setMem('lfHeight', - int(oPt.getMem('y') / 10 + 0.5))
      if this.bold
         oLog.setMem('lfWeight', FW_BOLD)
      endif
      if this.italic
         oLog.setMem('lfItalic', 255)
      endif
      if this.underline
         oLog.setMem('lfUnderline', 1)
      endif
      if this.Strikeout
         oLog.setMem('lfStrikeout', 1)
      endif
      oLog.setMem('lfFaceName', this.fontName + chr(0))
      oLog.setMem('lfCharSet', nChr)

      // restore the original dc.
      RMM_RestoreDC(nDc, -1)
      RMM_ReleaseDC(nHwnd, nDc)

      // save the structure for re-use.
      this.~logFont.string := oLog.string

      // destroy structures.
      release object oLog
      release object oPt
      oLog  := null
      oPt   := null

   return

   //------------------------------------------------------------//
   // place values into properties for dBL use.
   function x_Update
      this.bold      =  iif(this.~logFont.getMem('lfWeight') > FW_MEDIUM, true, false)
      this.italic    =  iif(this.~logFont.getMem('lfItalic') == 0, false, true)
      this.fontName  =  this.~logFont.getMem('lfFaceName')
      this.size      =  int(this.~chooseFont.getMem('iPointSize') / 10)
      if this.effects
         this.winColor  =  int(this.~chooseFont.getMem('rgbColors'))
         this.strikeout =  iif(this.~logFont.getMem('lfStrikeout') == 0, false, true)
         this.underline =  iif(this.~logFont.getMem('lfUnderline') == 0, false, true)
      else
         this.winColor  := 0
         this.strikeout := false
         this.underline := false
      endif
      class::x_Refresh()
      class::saveIni()
   return

   //------------------------------------------------------------//
   // write value to ini file.
   function x_WriteIni(cSect, cItem, cValu)
   return   this.~appIni.setValue(;
            cSect,;
            cItem,;
            cValu,;
            this.iniFile)

endclass

//---------------------------------------------------------------//
class FONTDLG_Structures
   //---------------------------------------------------------------//
   //------------------------------------------------------------//
   //       end of constructor   -  methods below
   //------------------------------------------------------------//
   function chooseFontStructure
      Local sStruct
      if not VSTRUCT_ISDEF("CHOOSEFONT")
         VSTRUCT_DEF('CHOOSEFONT',;
                  'lStructSize',       TYPE_DWORD,;
                  'hwndOwner',         TYPE_CHANDLE,;
                  'hDC',               TYPE_CHANDLE,;
                  'lpLogFont',         TYPE_CHANDLE,;
                  'iPointSize',        TYPE_INT,;
                  'Flags',             TYPE_DWORD,;
                  'rgbColors',         TYPE_DWORD,;
                  'lCustData',         TYPE_DWORD,;
                  'lpfnHook',          TYPE_POINTER,;
                  'lpTemplateName',    TYPE_POINTER,;
                  'hInstance',         TYPE_CHANDLE,;
                  'lpszStyle',         TYPE_POINTER,;
                  'nFontType',         TYPE_WORD,;
                  'MISSING_ALIGNMENT', TYPE_WORD,;
                  'nSizeMin',          TYPE_INT,;
                  'nSizeMax',          TYPE_INT)
      endif
      sStruct  =  new VStruct('CHOOSEFONT')
      sStruct.setMem('lStructSize', sStruct.getSize())
   return   sStruct

   //------------------------------------------------------------//
   function logFontStructure
      Local sStruct
      if not VSTRUCT_ISDEF("LOGFONT")
         VSTRUCT_DEF('LOGFONT',;
                  'lfHeight',          TYPE_LONG,;
                  'lfWidth',           TYPE_LONG,;
                  'lfEscapement',      TYPE_LONG,;
                  'lfOrientation',     TYPE_LONG,;
                  'lfWeight',          TYPE_LONG,;
                  'lfItalic',          TYPE_BYTE,;
                  'lfUnderline',       TYPE_BYTE,;
                  'lfStrikeOut',       TYPE_BYTE,;
                  'lfCharSet',         TYPE_BYTE,;
                  'lfOutPrecision',    TYPE_BYTE,;
                  'lfClipPrecision',   TYPE_BYTE,;
                  'lfQuality',         TYPE_BYTE,;
                  'lfPitchAndFamily',  TYPE_BYTE,;
                  'lfFaceName',        TYPE_STRING(32))
      endif
      sStruct  =  new VStruct('LOGFONT')
   return   sStruct

   //------------------------------------------------------------//
   function pointStructure
      Local sStruct
      if not VSTRUCT_ISDEF("POINT")
         VSTRUCT_DEF('POINT',;
                  'x',                 TYPE_LONG,;
                  'y',                 TYPE_LONG)
      endif
      sStruct  =  new VStruct('POINT')
   return   sStruct

endclass
//---------------------------------------------------------------//
//          end of FONTDLG_Structures//---------------------------------------------------------------//
